r"""Classes for the Wald Space and elements therein of class Wald and helper classes.

Lead author: Jonas Lueg

References
----------
.. [Garba21] Garba, M. K., T. M. W. Nye, J. Lueg and S. F. Huckemann.
    "Information geometry for phylogenetic trees"
    Journal of Mathematical Biology, 82(3):19, February 2021a.
    https://doi.org/10.1007/s00285-021-01553-x.
.. [Lueg21] Lueg, J., M. K. Garba, T. M. W. Nye, S. F. Huckemann.
    "Wald Space for Phylogenetic Trees."
    Geometric Science of Information, Lecture Notes in Computer Science,
    pages 710–717, Cham, 2021.
    https://doi.org/10.1007/978-3-030-80209-7_76.
"""

import random
from abc import ABC

import geomstats.backend as gs
from geomstats.geometry.hermitian_matrices import powermh
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.spd_matrices import SPDMatrices
from geomstats.geometry.stratified.point_set import (
    Point,
    PointBatch,
    PointSet,
    PointSetMetric,
)
from geomstats.geometry.stratified.trees import (
    ForestTopology,
    Split,
    delete_splits,
    generate_splits,
)
from geomstats.geometry.stratified.vectorization import (
    _manipulate_output,
    broadcast_lists,
    vectorize_point,
)
from geomstats.numerics.interpolation import LinearInterpolator1D
from geomstats.numerics.optimization import ScipyMinimize


def _manipulate_input_with_array(arg, name):
    if gs.is_array(arg):
        ndim = 2 if name == "ambient_point" else 1
        if arg.ndim > ndim:
            return arg, False
        return gs.expand_dims(arg, axis=0), True

    if not isinstance(arg, (list, tuple)):
        return [arg], True

    return arg, False


def _manipulate_output_wald(out, to_list):
    return _manipulate_output(out, to_list, manipulate_output_iterable=WaldBatch)


def make_splits(n_labels):
    """Generate all possible splits of a collection.

    Parameters
    ----------
    n_labels : int
        Number of labels.

    Returns
    -------
    split_iterator : Iterator[Split]
    """
    if n_labels <= 1:
        raise ValueError("`n_labels` must be greater than 1.")
    if n_labels == 2:
        yield Split(part1=[0], part2=[1])
    else:
        for split in make_splits(n_labels=n_labels - 1):
            yield Split(part1=split.part1, part2=split.part2.union((n_labels - 1,)))
            yield Split(part1=split.part1.union((n_labels - 1,)), part2=split.part2)
        yield Split(part1=list(range(n_labels - 1)), part2=[n_labels - 1])


def make_topologies(n_labels):
    """Generate all possible sets of compatible splits of a collection.

    This only works well for `len(n_labels) < 8`.

    Parameters
    ----------
    n_labels : int
        Number of labels.

    Returns
    -------
    topology_iterator : Iterator[ForestTopology]
    """
    if n_labels <= 1:
        raise ValueError("The collection must have 2 elements or more.")
    if n_labels in (2, 3):
        yield ForestTopology(
            partition=(tuple(range(n_labels)),),
            split_sets=(list(make_splits(n_labels)),),
        )
    else:
        pendant_split = Split(part1=[n_labels - 1], part2=list(range(n_labels - 1)))
        for st in make_topologies(n_labels - 1):
            for s in st.split_sets[0]:
                new_split_set = [pendant_split]
                a, b = set(s.part1), set(s.part2)
                for t in st.split_sets[0]:
                    _, d = set(t.part1), set(t.part2)
                    if t != s:
                        if a.issubset(d) or b.issubset(d):
                            new_split_set.append(
                                Split(
                                    part1=t.part1, part2=t.part2.union((n_labels - 1,))
                                )
                            )
                        else:
                            new_split_set.append(
                                Split(
                                    part1=t.part2, part2=t.part1.union((n_labels - 1,))
                                )
                            )
                    else:
                        new_split_set.append(
                            Split(part1=s.part1, part2=s.part2.union((n_labels - 1,)))
                        )
                        new_split_set.append(
                            Split(part1=s.part2, part2=s.part1.union((n_labels - 1,)))
                        )
                yield ForestTopology(
                    partition=(tuple(range(n_labels)),),
                    split_sets=(new_split_set,),
                )


def _generate_partition(n_labels, p_new):
    r"""Generate a random partition of :math:`\{0,\dots,n-1\}`.

    This algorithm works as follows: Start with a single set containing zero,
    then successively add the labels from 1 to n-1 to the partition in the
    following manner: for each label u, with probability `probability`, add the
    label u to a random existing set of the partition, else add a new singleton
    set {u} to the partition (i.e. with probability 1 - `probability`).

    Parameters
    ----------
    p_new : float
        A float between 0 and 1, the probability that no new component is added,
        and 1 - probability that a new component is added.

    Returns
    -------
    partition : list[list[int]]
        A partition of the set :math:`\{0,\dots,n-1\}` into non-empty sets.
    """
    _partition = [[0]]
    for u in range(1, n_labels):
        if gs.random.rand(1) < p_new:
            index = random.randint(0, len(_partition) - 1)
            _partition[index].append(u)
        else:
            _partition.append([u])
    return _partition


def generate_random_wald(n_labels, p_keep, p_new, btol=1e-8, check=True):
    """Generate a random instance of class ``Wald``.

    Parameters
    ----------
    n_labels : int
        The number of labels the wald is generated with respect to.
    p_keep : float
        The probability will be inserted into the generation of a partition as
        well as for the generation of a split set for the topology of the wald.
    p_new : float
        A float between 0 and 1, the probability that no new component is added,
        and probability of 1 - p_new_ that a new component is added.
    btol: float
        Tolerance for the boundary of the coordinates in each grove. Defaults to
        1e-08.
    check : bool
        If True, checks if splits still separate all labels. In this case, the split
        will not be deleted. If False, any split can be randomly deleted.

    Returns
    -------
    random_wald : Wald
        The randomly generated wald.
    """
    partition = _generate_partition(n_labels=n_labels, p_new=p_new)
    split_sets = [generate_splits(labels=_part) for _part in partition]

    split_sets = [
        delete_splits(splits=splits, labels=part, p_keep=p_keep, check=check)
        for part, splits in zip(partition, split_sets)
    ]

    top = ForestTopology(partition=partition, split_sets=split_sets)
    x = gs.random.uniform(size=(top.n_splits,), low=0, high=1)
    x = gs.minimum(gs.maximum(btol, x), 1 - btol)
    return Wald(topology=top, weights=x)


class Wald(Point):
    r"""A class for wälder, that are phylogenetic forests, elements of the Wald Space.

    A wald is essentially a phylogenetic forest with weights between zero and one on
    the edges. The forest structure is stored as a ``ForestTopology`` and the edge
    weights are an array of length that is equal to the total number of splits in the
    structure. These elements are the points in Wald space and other phylogenetic forest
    spaces, like BHV space, although the partition is just the whole set of labels in
    this case.

    Parameters
    ----------
    topology : ForestTopology
        The structure of the forest.
    weights : array-like, shape=[n_splits]
        The edge weights, array of floats between 0 and 1, with m entries, where m is
        the total number of splits/edges in the structure ``top``.
    corr : array-like, shape=[n_labels, n_labels]
        Correlation matrix of the topology with edge weights.
    """

    def __init__(self, topology, weights):
        super().__init__()
        self.topology = topology
        self.weights = weights

        self.corr = self.topology.corr(weights)

    def __repr__(self):
        """Return the string representation of the wald.

        This string representation requires that one can retrieve all necessary
        information from the string.

        Returns
        -------
        string_of_wald : str
            Return the string representation of the wald.
        """
        return repr((self.topology, tuple(self.weights)))

    def __str__(self):
        """Return the fancy printable string representation of the wald.

        This string representation does NOT require that one can retrieve all necessary
        information from the string, but this string representation is required to be
        readable by a human.

        Returns
        -------
        string_of_wald : str
            Return the fancy readable string representation of the wald.
        """
        return f"({str(self.topology)};{str(self.weights)})"

    def _equal_single(self, point, atol=gs.atol):
        """Check equality against another point.

        Parameters
        ----------
        point : Wald
            Point to compare against point.
        atol : float

        Returns
        -------
        is_equal : bool
        """
        if self.topology != point.topology:
            return False

        return gs.all(gs.abs(self.weights - point.weights) < atol)

    @vectorize_point((1, "point"))
    def equal(self, point, atol=gs.atol):
        """Check equality against another point.

        Parameters
        ----------
        point : Wald or WaldBatch
            Point to compare against point.
        atol : float

        Returns
        -------
        is_equal : array-like, shape=[...]
        """
        return gs.array([self._equal_single(point_, atol) for point_ in point])


class WaldBatch(PointBatch):
    """Wald batch."""

    @property
    def topology(self):
        """Forest topology.

        Returns
        -------
        topology : list[ForestTopology]
        """
        return [point.topology for point in self]

    @property
    def weights(self):
        """Edge weights.

        Returns
        -------
        weights : array-like, shape=[n_points, n_splits]
        """
        return gs.stack([point.weights for point in self])

    @property
    def corr(self):
        """Correlation matrix of the topology with edge weights.

        Returns
        -------
        corr : array-like, shape=[n_points, n_nodes, n_nodes]
        """
        return gs.stack([point.corr for point in self])


class WaldSpace(PointSet):
    r"""Class for the Wald space, a metric space for phylogenetic forests.

    A topological space. Points in Wald space are instances of the class :class:`Wald`:
    phylogenetic forests with edge weights between 0 and 1.
    In particular, Wald space is a stratified space, each stratum is called grove.
    The highest dimensional groves correspond to fully resolved or binary trees.
    The topology is obtained from embedding wälder into the ambient space of strictly
    positive :math:`n\times n` symmetric matrices, implemented in the
    class :class:`spd.SPDMatrices`.

    Parameters
    ----------
    n_labels : int
        Integer determining the number of labels in the forests, and thus the shape of
        the correlation matrices: n_labels x n_labels.

    Attributes
    ----------
    ambient_space : Manifold
        The ambient space, the positive definite n_labels x n_labels matrices that the
        WaldSpace is embedded into.
    """

    def __init__(self, n_labels, ambient_space=None, equip=True):
        super().__init__(equip)
        self.n_labels = n_labels

        if ambient_space is None:
            ambient_space = SPDMatrices(n=self.n_labels)
        self.ambient_space = ambient_space

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return WaldSpaceMetric

    def _belongs_single(self, point, atol=gs.atol):
        """Check if a point belongs to Tree space.

        Parameters
        ----------
        point : Wald
            The point to be checked.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : bool
            Boolean denoting if point belongs to wald space.
        """
        if not self.ambient_space.belongs(self.lift(point)):
            return False

        if gs.all(point.weights > 0) and gs.all(point.weights < 1):
            return True

        return False

    @vectorize_point((1, "point"))
    def belongs(self, point, atol=gs.atol):
        """Check if a point `wald` belongs to Wald space.

        From FUTURE PUBLICATION we know that the corresponding matrix of a wald is
        strictly positive definite if and only if the labels are separated by at least
        one edge, which is exactly when the wald is an element of the Wald space.

        Parameters
        ----------
        point : Wald or WaldBatch
            The point to be checked.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...]
            Boolean denoting if `point` belongs to Wald space.
        """
        return gs.array([self._belongs_single(point_, atol) for point_ in point])

    def random_point(self, n_samples=1, p_tree=0.9, p_keep=0.9, btol=1e-8):
        """Sample a random point in Wald space.

        Parameters
        ----------
        n_samples : int
            Number of samples. Defaults to 1.
        p_tree : float between 0 and 1
            The probability that the sampled point is a tree, and not a forest. If the
            probability is equal to 1, then the sampled point will be a tree.
            Defaults to 0.9.
        p_keep : float between 0 and 1
            The probability that a sampled edge is kept and not deleted randomly.
            To be precise, it is not exactly the probability, as some edges cannot be
            deleted since the requirement that two labels are separated by a split might
            be violated otherwise.
            Defaults to 0.9
        btol: float
            Tolerance for the boundary of the coordinates in each grove. Defaults to
            1e-08.

        Returns
        -------
        samples : Wald or WaldBatch
            Points sampled in Wald space.
        """
        p_new = p_tree ** (1 / (self.n_labels - 1))
        forests = [
            generate_random_wald(self.n_labels, p_keep, p_new, btol, check=True)
            for _ in range(n_samples)
        ]

        if n_samples == 1:
            return forests[0]

        return WaldBatch(forests)

    def random_grove_point(self, topology, n_samples=1):
        """Sample a random point in a given grove of wald spcae.

        Parameters
        ----------
        topology : ForestTopology
        n_samples : int
            Number of samples. Defaults to 1.

        Returns
        -------
        samples : Wald or WaldBatch
            Points sampled in Wald space.
        """
        n_splits = topology.n_splits
        weights = gs.random.uniform(size=(n_samples, n_splits))

        forests = [Wald(topology, weights_) for weights_ in weights]

        if n_samples == 1:
            return forests[0]

        return WaldBatch(forests)

    def lift(self, point):
        """Lift a point to the ambient space.

        Parameters
        ----------
        point : Wald or WaldBatch
            The point to be lifted.

        Returns
        -------
        lifted_point : array-like, shape=[..., n_labels, n_labels]
            Point in the ambient space.
        """
        return point.corr


class WaldSpaceMetric(PointSetMetric):
    """Wald space metric.

    Parameters
    ----------
    space : WaldSpace
        Set to equip with metric.
    projection_solver : ProjectionSolver
        Numerical solver to solve projection problem.
    geodesic_solver : GeodesicSolver
        Numerical solver to solve geodesic problem.
    """

    def __init__(self, space, projection_solver=None, geodesic_solver=None):
        super().__init__(space)

        if projection_solver is None:
            projection_solver = LocalProjectionSolver(space)

        if geodesic_solver is None:
            geodesic_solver = SuccessiveProjectionGeodesicSolver(space)

        self.projection_solver = projection_solver
        self.geodesic_solver = geodesic_solver

    def dist(self, point_a, point_b):
        """Distance between two points in the WaldSpace.

        Parameters
        ----------
        point_a: Wald or WaldBatch
            Point in the WaldSpace.
        point_b: Wald or WaldBatch
            Point in the WaldSpace.

        Returns
        -------
        distance : array-like, shape=[...]
            Distance.
        """
        discrete_path = self.discrete_geodesic(point_a, point_b)
        is_list = not isinstance(discrete_path, WaldBatch)
        if not is_list:
            discrete_path = [discrete_path]

        dists = [
            gs.sum(
                self._space.ambient_space.metric.dist(
                    discrete_path_.corr[1:], discrete_path_.corr[:-1]
                ),
                axis=-1,
            )
            for discrete_path_ in discrete_path
        ]

        if is_list:
            return gs.stack(dists)

        return dists[0]

    def geodesic(self, initial_point, end_point):
        """Compute the geodesic in the WaldSpace.

        Parameters
        ----------
        initial_point: Wald or WaldBatch
            Point in the WaldSpace.
        end_point: Wald or WaldBatch
            Point in the WaldSpace.

        Returns
        -------
        path : callable
            Time parameterized geodesic curve.
        """
        return self.geodesic_solver.geodesic(initial_point, end_point)

    def discrete_geodesic(self, initial_point, end_point):
        """Compute a discrete geodesic in the WaldSpace.

        Parameters
        ----------
        initial_point: Wald or WaldBatch
            Point in the WaldSpace.
        end_point: Wald or WaldBatch
            Point in the WaldSpace.

        Returns
        -------
        geod_points : WaldBatch or list[WaldBatch]
            Time parameterized geodesic curve.
        """
        return self.geodesic_solver.discrete_geodesic(initial_point, end_point)

    def projection(self, ambient_point, topology, initial_weights=None):
        """Projects a point into Wald space.

        Parameters
        ----------
        ambient_point : array-like, shape=[..., n_nodes, n_nodes]
            Ambient point to project.
        topology : ForestTopology or list[ForestTopology]
            Stratum topology.
        initial_weights : array-like, shape=[..., n_splits]
            Initial guess for weights.
        """
        return self.projection_solver.projection(
            ambient_point, topology, initial_weights=initial_weights
        )


def _squared_dist_and_grad_affine(space, topology, ambient_point):
    """Squared distance and gradient wrt weights.

    See section 5.1 of [Garba2021]_.

    Parameters
    ----------
    space : WaldSpace
    topology : ForestTopology
    ambient_point : array-like, shape=[n_nodes, n_nodes]
        Point wrt measure distance.

    Returns
    -------
    value_and_grad : callable
        A callable that takes weights and outputs value and grad.
    """
    sqrt_ambient_point, inv_sqrt_ambient_point = powermh(
        ambient_point, [1.0 / 2, -1.0 / 2]
    )

    def _value_and_grad(weights):
        corr = topology.corr(weights)
        inv_corr = gs.linalg.inv(corr)
        grad = topology.corr_gradient(weights)

        target = space.ambient_space.metric.squared_dist(corr, ambient_point)

        target_grad = 2 * gs.trace(
            Matrices.mul(
                gs.linalg.logm(
                    Matrices.mul(inv_sqrt_ambient_point, corr, inv_sqrt_ambient_point)
                ),
                sqrt_ambient_point,
                inv_corr,
                grad,
                inv_sqrt_ambient_point,
            )
        )

        return target, target_grad

    return _value_and_grad


def _squared_dist_and_grad_euclidean(space, topology, ambient_point):
    """Squared distance and gradient wrt weights.

    Parameters
    ----------
    space : WaldSpace
    topology : ForestTopology
    ambient_point : array-like, shape=[n_nodes, n_nodes]
        Point wrt measure distance.

    Returns
    -------
    value_and_grad : callable
        A callable that takes weights and outputs value and grad.
    """

    def _value_and_grad(weights):
        corr = topology.corr(weights)
        grad = topology.corr_gradient(weights)

        target = space.ambient_space.metric.squared_dist(corr, ambient_point)
        target_grad = 2 * gs.sum((corr - ambient_point) * grad, axis=(-2, -1))
        return target, target_grad

    return _value_and_grad


def _squared_dist_and_grad_autodiff(space, topology, ambient_point):
    """Squared distance and gradient wrt weights.

    Parameters
    ----------
    space : WaldSpace
    topology : ForestTopology
    ambient_point : array-like, shape=[n_nodes, n_nodes]
        Point wrt measure distance.

    Returns
    -------
    value_and_grad : callable
        A callable that takes weights and outputs value and grad.
    """

    def _value(weights):
        corr = topology.corr(weights)
        return space.ambient_space.metric.squared_dist(corr, ambient_point)

    return gs.autodiff.value_and_grad(_value)


_AMBIENT_METRIC_TO_SQUARED_DIST_GRAD = {
    "SPDAffineMetric": _squared_dist_and_grad_affine,
    "SPDEuclideanMetric": _squared_dist_and_grad_euclidean,
}


class LocalProjectionSolver:
    """Local projection solver."""

    def __init__(self, space, btol=1e-10):
        self._space = space
        self.btol = btol
        self.optimizer = ScipyMinimize(
            method="L-BFGS-B",
            tol=gs.atol,
        )

    def _get_bounds(self, n_splits):
        return [(self.btol, 1 - self.btol)] * n_splits

    def _projection_single(self, ambient_point, topology, initial_weights=None):
        """Project ambient point into wald space.

        Parameters
        ----------
        ambient_point : array-like, shape=[n_nodes, n_nodes]
            Ambient point to project.
        topology : ForestTopology
            Stratum topology.
        initial_weights : array-like, shape=[n_splits]
            Initial guess for weights.
        """
        if len(topology.partition) == topology.n_labels:
            return Wald(topology=topology, weights=gs.ones(self.n_labels))

        value_and_grad = _AMBIENT_METRIC_TO_SQUARED_DIST_GRAD.get(
            self._space.ambient_space.metric.__class__.__name__,
            _squared_dist_and_grad_autodiff,
        )(
            space=self._space,
            ambient_point=ambient_point,
            topology=topology,
        )

        n_splits = topology.n_splits

        initial_weights = (
            gs.ones(n_splits) * 0.5 if initial_weights is None else initial_weights
        )

        self.optimizer.bounds = self._get_bounds(n_splits)
        res = self.optimizer.minimize(value_and_grad, initial_weights, fun_jac=True)

        if res.status != 0:
            raise ValueError("Projection failed!")

        weights = gs.array(
            [
                _x if self.btol < _x < 1 - self.btol else 0 if _x <= self.btol else 1
                for _x in res.x
            ]
        )

        return Wald(topology=topology, weights=weights)

    @vectorize_point(
        (1, "ambient_point"),
        (2, "topology"),
        (3, "initial_weights"),
        manipulate_input=_manipulate_input_with_array,
        manipulate_output=_manipulate_output_wald,
    )
    def projection(self, ambient_point, topology, initial_weights=None):
        """Project ambient point into wald space.

        Parameters
        ----------
        ambient_point : array-like, shape=[..., n_nodes, n_nodes]
            Ambient point to project.
        topology : ForestTopology or list[ForestTopology]
            Stratum topology.
        initial_weights : array-like, shape=[..., n_splits]
            Initial guess for weights.
        """
        ambient_point, topology, initial_weights = broadcast_lists(
            ambient_point, topology, initial_weights
        )
        return [
            self._projection_single(ambient_point_, topology_, initial_weights_)
            for ambient_point_, topology_, initial_weights_ in zip(
                ambient_point, topology, initial_weights
            )
        ]


class BasicWaldGeodesicSolver(ABC):
    """Abstract class for wald geodesic solver.

    Parameters
    ----------
    space : WaldSpace
    """

    def __init__(self, space):
        self._space = space

    @vectorize_point(
        (1, "initial_point"),
        (2, "end_point"),
        manipulate_output=lambda out, to_list: _manipulate_output(
            out, to_list, manipulate_output_iterable=lambda x: x
        ),
    )
    def discrete_geodesic(self, initial_point, end_point):
        """Compute a discrete geodesic in the WaldSpace.

        Parameters
        ----------
        initial_point: Wald or WaldBatch
            Point in the WaldSpace.
        initial_point: Wald or WaldBatch
            Point in the WaldSpace.

        Returns
        -------
        geod_points : WaldBatch or list[WaldBatch]
            Time parameterized geodesic curve.
        """
        initial_point, end_point = broadcast_lists(initial_point, end_point)
        return [
            self._discrete_geodesic_single(
                initial_point_,
                end_point_,
            )
            for initial_point_, end_point_ in zip(initial_point, end_point)
        ]

    def geodesic(self, initial_point, end_point):
        """Compute the geodesic in the WaldSpace.

        Parameters
        ----------
        initial_point: Wald or WaldBatch
            Point in the WaldSpace.
        initial_point: Wald or WaldBatch
            Point in the WaldSpace.

        Returns
        -------
        path : callable
            Time parameterized geodesic curve.
        """

        def _vec(t, fncs):
            if len(fncs) == 1:
                return fncs[0](t)

            return [fnc(t) for fnc in fncs]

        discrete_path = self.discrete_geodesic(initial_point, end_point)

        if isinstance(discrete_path, WaldBatch):
            return DiscreteWaldPath(self._space, discrete_path)

        return lambda t: _vec(
            t,
            fncs=[
                DiscreteWaldPath(self._space, discrete_path_)
                for discrete_path_ in discrete_path
            ],
        )


class NaiveProjectionGeodesicSolver(BasicWaldGeodesicSolver):
    """Naive geodesic projection solver.

    Implementation of algorithm 1 from [Lueg21]_.
    """

    def __init__(self, space, n_grid=10):
        super().__init__(space)
        self.n_grid = n_grid

    def _discrete_geodesic_single(self, initial_point, end_point):
        """Compute a discrete geodesic in the WaldSpace.

        Parameters
        ----------
        initial_point: Wald
            Point in the WaldSpace.
        end_point: Wald
            Point in the WaldSpace.

        Returns
        -------
        geod_points : WaldBatch
            Time parameterized geodesic curve.
        """
        if initial_point.topology != end_point.topology:
            raise ValueError("Can only handle points in the same grove.")

        topology = initial_point.topology

        initial_corr = self._space.lift(initial_point)
        end_corr = self._space.lift(end_point)

        ambient_geod_func = self._space.ambient_space.metric.geodesic(
            initial_corr, end_point=end_corr
        )

        time = gs.linspace(0, 1, self.n_grid)[1:-1]
        mid_ambient_point = ambient_geod_func(time)

        mid_points = self._space.metric.projection(mid_ambient_point, topology)

        return WaldBatch([initial_point] + mid_points + [end_point])


class SuccessiveProjectionGeodesicSolver(BasicWaldGeodesicSolver):
    """Successive projection geodesic projection solver.

    Implementation of algorithm 2 from [Lueg21]_.
    """

    def __init__(self, space, n_grid=10):
        super().__init__(space)
        self.n_grid = n_grid

    def _discrete_geodesic_single(self, initial_point, end_point):
        """Compute a discrete geodesic in the WaldSpace.

        Parameters
        ----------
        initial_point: Wald
            Point in the WaldSpace.
        end_point: Wald
            Point in the WaldSpace.

        Returns
        -------
        geod_points : WaldBatch
            Time parameterized geodesic curve.
        """
        if initial_point.topology != end_point.topology:
            raise ValueError("Can only handle points in the same grove.")

        topology = initial_point.topology

        left_points = [initial_point]
        right_points = [end_point]
        for i in range(2, self.n_grid // 2 + 1):
            time = 1 / (self.n_grid - i + 1)

            left_corr = self._space.lift(left_points[-1])
            right_corr = self._space.lift(right_points[-1])

            ambient_geod_func = self._space.ambient_space.metric.geodesic(
                left_corr, right_corr
            )

            left_points.append(
                self._space.metric.projection(
                    ambient_geod_func(time)[0],
                    topology,
                    initial_weights=left_points[-1].weights,
                )
            )
            right_points.append(
                self._space.metric.projection(
                    ambient_geod_func(1 - time)[0],
                    topology,
                    initial_weights=right_points[-1].weights,
                )
            )

        right_points.reverse()

        if gs.mod(self.n_grid, 2) == 0:
            return WaldBatch(left_points + right_points)

        left_corr = self._space.lift(left_points[-1])
        right_corr = self._space.lift(right_points[0])

        ambient_geod_func = self._space.ambient_space.metric.geodesic(
            left_corr, end_point=right_corr
        )

        mid_point = self._space.metric.projection(ambient_geod_func(0.5), topology)
        return WaldBatch(left_points + mid_point + right_points)


class DiscreteWaldPath:
    """A uniformly-sampled discrete path.

    Parameters
    ----------
    space : WaldSpace
    path : WaldBatch
        Wald collection with common topology.
    interpolator : Interpolator1D
    """

    def __init__(self, space, path, interpolator=None, **interpolator_kwargs):
        # NB: assumes common topology
        self._topology = path[0].topology

        times = self._get_times(space, path)

        if interpolator is None:
            interpolator = LinearInterpolator1D(
                times, path.weights, point_ndim=1, **interpolator_kwargs
            )
        self.interpolator = interpolator

    def _get_times(self, space, path):
        """Compute times for linear interpolation."""
        dists = space.ambient_space.metric.dist(path[1:].corr, path[:-1].corr)
        cum_dists = gs.cumsum(dists)
        return gs.concatenate([gs.array([0.0]), cum_dists / cum_dists[-1]])

    def __call__(self, t):
        """Interpolate path.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        path : WaldBatch
        """
        weights = self.interpolator(t)
        return WaldBatch([Wald(self._topology, weights_) for weights_ in weights])



import pytest

import geomstats.backend as gs
from geomstats.test_cases.geometry.pullback_metric import PullbackDiffeoMetricTestCase


def univariate_fisher_rao_dist(point_a, point_b):
    mean_a = point_a[..., 0]
    mean_b = point_b[..., 0]
    sigma_a = point_a[..., 1]
    sigma_b = point_b[..., 1]

    sdiff_mean = (mean_a - mean_b) ** 2
    sdiff_sigma = (sigma_a - sigma_b) ** 2
    ssum_sigma = (sigma_a + sigma_b) ** 2

    return (
        2
        * gs.sqrt(2)
        * gs.arctanh(
            gs.sqrt((sdiff_mean + 2 * sdiff_sigma) / (sdiff_mean + 2 * ssum_sigma))
        )
    )


class UnivariateNormalMetricTestCase(PullbackDiffeoMetricTestCase):
    @pytest.mark.random
    def test_dist_against_closed_form(self, n_points, atol):
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        res = self.space.metric.dist(point_a, point_b)
        expected = univariate_fisher_rao_dist(point_a, point_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_sectional_curvature_value(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        res = self.space.metric.sectional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertTrue(gs.all(gs.abs(res + 1 / 2) < atol))



"""Decorator to handle vectorization."""

import functools

import geomstats.backend as gs
from geomstats.geometry.stratified.point_set import PointBatch


def broadcast_lists(*lists):
    """Broadcast lists.

    Similar behavior as ``gs.broadcast_arrays``, but for lists.

    Parameters
    ----------
    *lists : list

    Returns
    -------
    *broadcasted_lists : list
        Lists with broadcasted length.
    """
    lens = [len(list_) for list_ in lists]
    n_max = max(lens)
    if n_max == 1:
        return lists

    out = []
    for len_, list_ in zip(lens, lists):
        if len_ == 1:
            out.append([list_[0]] * n_max)

        elif len_ == n_max:
            out.append(list_)

        else:
            raise Exception(f"Cannot broadcast lens: {lens}")

    return out


def _manipulate_input(arg, *args):
    """Transform input into list if not one.

    Parameters
    ----------
    arg : any
        Argument to manipulate

    Returns
    -------
    arg : any or list[any]
    transformed : bool
        If point was transformed.
    """
    if not isinstance(arg, (list, tuple)):
        return [arg], True

    return arg, False


def _manipulate_output_iterable(out):
    return PointBatch(out)


def _manipulate_output(
    out, to_list, manipulate_output_iterable=_manipulate_output_iterable
):
    is_array = gs.is_array(out)
    is_iterable = isinstance(out, (list, tuple))

    if not (gs.is_array(out) or is_iterable):
        return out

    if to_list:
        if is_array:
            return gs.array(out[0])
        if is_iterable:
            return out[0]

    if is_iterable:
        return manipulate_output_iterable(out)

    return out


def vectorize_point(
    *args_positions,
    manipulate_input=_manipulate_input,
    manipulate_output=_manipulate_output,
):
    """Check point type and transform in iterable if not the case.

    Parameters
    ----------
    args_positions : tuple
        Position and corresponding argument name. A tuple for each position.

    Notes
    -----
    Explicitly defining args_positions and args names ensures it works for all
    combinations of input calling.
    """

    def _dec(func):
        @functools.wraps(func)
        def _wrapped(*args, **kwargs):
            to_list = True
            args = list(args)
            for pos, name in args_positions:
                if name in kwargs:
                    kwargs[name], to_list_ = manipulate_input(kwargs[name], name)
                else:
                    args[pos], to_list_ = manipulate_input(args[pos], name)

                to_list = to_list and to_list_

            out = func(*args, **kwargs)

            return manipulate_output(out, to_list)

        return _wrapped

    return _dec



"""Autograd based random backend."""

import autograd.numpy as _np
from autograd.numpy.random import randint, seed

from .._shared_numpy.random import choice, multivariate_normal, normal, rand, uniform



import random

import pytest
import scipy

import geomstats.backend as gs
from geomstats.learning.frechet_mean import FrechetMean
from geomstats.test.random import (
    EmbeddedSpaceRandomDataGenerator,
    HypersphereIntrinsicRandomDataGenerator,
)
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import LevelSetTestCase
from geomstats.test_cases.geometry.manifold import ManifoldTestCase
from geomstats.vectorization import get_batch_shape


def _estimate_von_mises_kappa(X, n_steps=100):
    n_points = X.shape[0]
    dim = X.shape[1] - 1

    sum_points = gs.sum(X, axis=0)
    mean_norm = gs.linalg.norm(sum_points) / n_points

    kappa_estimate = mean_norm * (dim + 1.0 - mean_norm**2) / (1.0 - mean_norm**2)
    kappa_estimate = gs.cast(kappa_estimate, gs.float64)
    p = dim + 1
    for _ in range(n_steps):
        bessel_func_1 = scipy.special.iv(p / 2.0, kappa_estimate)
        bessel_func_2 = scipy.special.iv(p / 2.0 - 1.0, kappa_estimate)
        ratio = bessel_func_1 / bessel_func_2
        denominator = 1.0 - ratio**2 - (p - 1.0) * ratio / kappa_estimate
        kappa_estimate -= (ratio - mean_norm) / denominator

    return kappa_estimate


def _belongs_intrinsic(space, point, atol=gs.atol):
    # TODO: use this concept in other places
    shape = point.shape[: -space.point_ndim]

    if point.shape[-1] == space.dim:
        return gs.ones(shape, dtype=bool)

    return gs.zeros(shape, dtype=bool)


def _is_tangent_intrinsic(space, tangent_vec, point, atol=gs.atol):
    shape = get_batch_shape(space.point_ndim, point, tangent_vec)
    if tangent_vec.shape[-1] == space.dim:
        return gs.ones(shape, dtype=bool)

    return gs.zeros(shape, dtype=bool)


class HypersphereCoordsTransformTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.extrinsic_data_generator = EmbeddedSpaceRandomDataGenerator(self.space)
            self.intrinsic_data_generator = HypersphereIntrinsicRandomDataGenerator(
                self.space_intrinsic
            )

    def _test_belongs_intrinsic(self, point, expected):
        res = _belongs_intrinsic(self.space_intrinsic, point)
        self.assertAllEqual(res, expected)

    def test_intrinsic_to_extrinsic_coords(self, point_intrinsic, expected, atol):
        res = self.space.intrinsic_to_extrinsic_coords(point_intrinsic)
        self.assertAllClose(res, expected, atol=atol)
        self.assertEqual(
            point_intrinsic.ndim,
            res.ndim,
            msg=f"`point_intrinsic.shape` is {point_intrinsic.shape} and "
            f"`point_extrinsic.shape` is {res.shape}",
        )

    @pytest.mark.vec
    def test_intrinsic_to_extrinsic_coords_vec(self, n_reps, atol):
        point_intrinsic = self.intrinsic_data_generator.random_point()

        try:
            expected = self.space.intrinsic_to_extrinsic_coords(point_intrinsic)
        except NotImplementedError:
            return

        vec_data = generate_vectorization_data(
            data=[dict(point_intrinsic=point_intrinsic, expected=expected, atol=atol)],
            arg_names=["point_intrinsic"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_intrinsic_to_extrinsic_coords_belongs(self, n_points, atol):
        point_intrinsic = self.intrinsic_data_generator.random_point(n_points)

        try:
            point_extrinsic = self.space.intrinsic_to_extrinsic_coords(point_intrinsic)
        except NotImplementedError:
            return

        res = self.space_extrinsic.belongs(point_extrinsic, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    def test_extrinsic_to_intrinsic_coords(self, point_extrinsic, expected, atol):
        res = self.space.extrinsic_to_intrinsic_coords(point_extrinsic)
        self.assertAllClose(res, expected, atol=atol)
        self.assertEqual(
            point_extrinsic.ndim,
            res.ndim,
            msg=f"`point_extrinsic.shape` is {point_extrinsic.shape} and "
            f"`point_intrinsic.shape` is {res.shape}",
        )

    @pytest.mark.vec
    def test_extrinsic_to_intrinsic_coords_vec(self, n_reps, atol):
        point_extrinsic = self.extrinsic_data_generator.random_point()

        try:
            expected = self.space.extrinsic_to_intrinsic_coords(point_extrinsic)
        except NotImplementedError:
            return

        vec_data = generate_vectorization_data(
            data=[dict(point_extrinsic=point_extrinsic, expected=expected, atol=atol)],
            arg_names=["point_extrinsic"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_extrinsic_to_intrinsic_coords_belongs(self, n_points):
        point_extrinsic = self.extrinsic_data_generator.random_point(n_points)

        try:
            point_intrinsic = self.space.extrinsic_to_intrinsic_coords(point_extrinsic)
        except NotImplementedError:
            return

        expected = gs.ones(n_points, dtype=bool)
        self._test_belongs_intrinsic(point_intrinsic, expected)

    @pytest.mark.random
    def test_intrinsic_to_extrinsic_coords_after_extrinsic_to_intrinsic(
        self, n_points, atol
    ):
        point_extrinsic = self.extrinsic_data_generator.random_point(n_points)

        try:
            point_intrinsic = self.space.extrinsic_to_intrinsic_coords(point_extrinsic)
        except NotImplementedError:
            return

        point_extrinsic_ = self.space.intrinsic_to_extrinsic_coords(point_intrinsic)

        self.assertAllClose(point_extrinsic_, point_extrinsic, atol=atol)

    @pytest.mark.random
    def test_extrinsic_to_intrinsic_coords_after_intrinsic_to_extrinsic_coords(
        self, n_points, atol
    ):
        point_intrinsic = self.intrinsic_data_generator.random_point(n_points)

        try:
            point_extrinsic = self.space.intrinsic_to_extrinsic_coords(point_intrinsic)
        except NotImplementedError:
            return

        point_intrinsic_ = self.space.extrinsic_to_intrinsic_coords(point_extrinsic)

        self.assertAllClose(point_intrinsic_, point_intrinsic, atol=atol)

    def test_tangent_spherical_to_extrinsic(
        self,
        tangent_vec_spherical,
        base_point_spherical,
        expected,
        atol,
    ):
        res = self.space.tangent_spherical_to_extrinsic(
            tangent_vec_spherical, base_point_spherical
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tangent_spherical_to_extrinsic_vec(self, n_reps, atol):
        base_point_spherical = self.intrinsic_data_generator.random_point()
        tangent_vec_spherical = self.intrinsic_data_generator.random_tangent_vec(
            base_point_spherical
        )

        try:
            expected = self.space.tangent_spherical_to_extrinsic(
                tangent_vec_spherical, base_point_spherical
            )
        except NotImplementedError:
            return

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_spherical=tangent_vec_spherical,
                    base_point_spherical=base_point_spherical,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_spherical", "base_point_spherical"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tangent_spherical_to_extrinsic_is_tangent(self, n_points, atol):
        base_point_spherical = self.intrinsic_data_generator.random_point(n_points)
        tangent_vec_spherical = self.intrinsic_data_generator.random_tangent_vec(
            base_point_spherical
        )

        try:
            tangent_vec = self.space.tangent_spherical_to_extrinsic(
                tangent_vec_spherical,
                base_point_spherical,
            )
        except NotImplementedError:
            return

        base_point = self.space.intrinsic_to_extrinsic_coords(base_point_spherical)

        res = self.space_extrinsic.is_tangent(tangent_vec, base_point, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    def test_tangent_extrinsic_to_spherical(
        self, tangent_vec, base_point, expected, atol
    ):
        res = self.space.tangent_extrinsic_to_spherical(
            tangent_vec,
            base_point=base_point,
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tangent_extrinsic_to_spherical_vec(self, n_reps, atol):
        base_point = self.extrinsic_data_generator.random_point()
        tangent_vec = self.extrinsic_data_generator.random_tangent_vec(base_point)

        try:
            expected = self.space.tangent_extrinsic_to_spherical(
                tangent_vec,
                base_point=base_point,
            )
        except NotImplementedError:
            return

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tangent_extrinsic_to_spherical_is_tangent(self, n_points):
        base_point = self.extrinsic_data_generator.random_point(n_points)
        tangent_vec = self.extrinsic_data_generator.random_tangent_vec(base_point)

        try:
            tangent_vec_spherical = self.space.tangent_extrinsic_to_spherical(
                tangent_vec,
                base_point=base_point,
            )
        except NotImplementedError:
            return

        base_point_spherical = self.space.extrinsic_to_intrinsic_coords(base_point)

        res = _is_tangent_intrinsic(
            self.space_intrinsic, tangent_vec_spherical, base_point_spherical
        )
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_tangent_extrinsic_to_spherical_after_tangent_spherical_to_extrinsic(
        self, n_points, atol
    ):
        base_point_spherical = self.intrinsic_data_generator.random_point(n_points)
        tangent_vec_spherical = self.intrinsic_data_generator.random_tangent_vec(
            base_point_spherical
        )

        try:
            tangent_vec = self.space.tangent_spherical_to_extrinsic(
                tangent_vec_spherical, base_point_spherical=base_point_spherical
            )
        except NotImplementedError:
            return

        base_point = self.space.intrinsic_to_extrinsic_coords(base_point_spherical)

        tangent_vec_spherical_ = self.space.tangent_extrinsic_to_spherical(
            tangent_vec, base_point
        )

        self.assertAllClose(tangent_vec_spherical_, tangent_vec_spherical, atol=atol)

    @pytest.mark.random
    def test_tangent_spherical_to_extrinsic_after_tangent_extrinsic_to_spherical(
        self, n_points, atol
    ):
        base_point = self.extrinsic_data_generator.random_point(n_points)
        tangent_vec = self.extrinsic_data_generator.random_tangent_vec(base_point)

        try:
            tangent_vec_spherical = self.space.tangent_extrinsic_to_spherical(
                tangent_vec, base_point
            )
        except NotImplementedError:
            return

        base_point_spherical = self.space.extrinsic_to_intrinsic_coords(base_point)

        tangent_vec_ = self.space.tangent_spherical_to_extrinsic(
            tangent_vec_spherical, base_point_spherical=base_point_spherical
        )

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)


class HypersphereExtrinsicTestCase(LevelSetTestCase):
    def _get_random_kappa(self, size=1):
        sample = gs.random.uniform(low=1.0, high=10000.0, size=(size,))
        if size == 1:
            return sample[0]
        return sample

    def _get_random_precision(self, precision_type="array"):
        if precision_type is None:
            return None

        precision = gs.random.uniform(low=1.0, high=10.0, size=(1,))[0]
        if precision_type is float:
            return precision

        return precision * gs.eye(self.space.dim)

    def _fit_frechet_mean(self, sample):
        if self.space.dim == 1:
            estimator = FrechetMean(self.space)
        else:
            estimator = FrechetMean(self.space, method="adaptive")

        estimator.fit(sample)

        return estimator.estimate_

    @pytest.mark.random
    def test_replace_values(self, n_points, atol):
        points = self.data_generator.random_point(n_points)
        new_points = self.data_generator.random_point(n_points)

        n_indices = random.randint(1, n_points)
        indices_int = [
            (index,) for index in random.sample(range(0, n_points), n_indices)
        ]
        if n_indices:
            indices = gs.array_from_sparse(
                indices_int, [True for _ in range(n_indices)], (n_points,)
            )
        else:
            indices = gs.zeros(n_points, dtype=bool)

        replaced_points = self.space._replace_values(
            points, new_points[indices], indices
        )

        self.assertAllClose(replaced_points[indices], new_points[indices], atol=atol)

    @pytest.mark.random
    def test_random_von_mises_fisher_sample_mean(
        self, n_samples, atol, random_mu=True, max_iter=100
    ):
        if random_mu:
            mu = self.space.random_point()
            expected = mu
        else:
            mu = None
            expected = gs.array([1.0] + [0.0] * self.space.dim)

        kappa = self._get_random_kappa()

        try:
            point = self.space.random_von_mises_fisher(
                n_samples=n_samples, mu=mu, kappa=kappa, max_iter=max_iter
            )
        except NotImplementedError:
            return

        sum_point = gs.sum(point, axis=0)
        res = sum_point / gs.linalg.norm(sum_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.validation
    def test_random_von_mises_fisher_sample_kappa(
        self,
        n_samples,
        atol,
        max_iter=100,
    ):
        kappa = 1.0

        try:
            X = self.space.random_von_mises_fisher(
                n_samples=n_samples, kappa=kappa, max_iter=max_iter
            )
        except NotImplementedError:
            return

        kappa_estimate = _estimate_von_mises_kappa(X, n_steps=100)
        self.assertAllClose(kappa_estimate, gs.array(kappa), atol=atol)

    @pytest.mark.random
    def test_random_von_mises_fisher_belongs(
        self, n_points, random_mu, atol, max_iter=100
    ):
        mu = self.space.random_point() if random_mu else None
        kappa = self._get_random_kappa()

        try:
            point = self.space.random_von_mises_fisher(
                n_samples=n_points, mu=mu, kappa=kappa, max_iter=max_iter
            )
        except NotImplementedError:
            return

        expected = gs.ones(n_points, dtype=bool)
        self.test_belongs(point, expected, atol)

    @pytest.mark.shape
    def test_random_von_mises_fisher_shape(self, n_points, random_mu, max_iter=100):
        mu = self.space.random_point() if random_mu else None
        kappa = self._get_random_kappa()

        try:
            point = self.space.random_von_mises_fisher(
                n_samples=n_points, mu=mu, kappa=kappa, max_iter=max_iter
            )
        except NotImplementedError:
            return

        expected_ndim = self.space.point_ndim + int(n_points > 1)
        self.assertEqual(gs.ndim(point), expected_ndim)

        self.assertAllEqual(gs.shape(point)[-self.space.point_ndim :], self.space.shape)

        if n_points > 1:
            self.assertEqual(gs.shape(point)[0], n_points)

    @pytest.mark.random
    def test_random_riemannian_normal_belongs(
        self, n_samples, random_mean, precision_type, atol, max_iter=100
    ):
        mean = self.space.random_point() if random_mean else None
        precision = self._get_random_precision(precision_type)

        sample = self.space.random_riemannian_normal(
            mean=mean, precision=precision, n_samples=n_samples, max_iter=max_iter
        )

        expected = gs.ones(n_samples, dtype=bool)
        self.test_belongs(sample, expected, atol)

    @pytest.mark.shape
    def test_random_riemannian_normal_shape(
        self, n_samples, random_mean, precision_type, atol, max_iter=100
    ):
        mean = self.space.random_point() if random_mean else None
        precision = self._get_random_precision(precision_type)

        point = self.space.random_riemannian_normal(
            mean=mean, precision=precision, n_samples=n_samples, max_iter=max_iter
        )

        expected_ndim = self.space.point_ndim + int(n_samples > 1)
        self.assertEqual(gs.ndim(point), expected_ndim)

        self.assertAllEqual(gs.shape(point)[-self.space.point_ndim :], self.space.shape)

        if n_samples > 1:
            self.assertEqual(gs.shape(point)[0], n_samples)

    @pytest.mark.random
    def test_random_riemannian_normal_frechet_mean(
        self, n_samples, random_mean, atol, max_iter=100
    ):
        if random_mean:
            expected = mean = self.space.random_point()
        else:
            mean = None
            expected = gs.array([0.0] * self.space.dim + [1.0])  # north pole

        precision = self._get_random_precision()

        sample = self.space.random_riemannian_normal(
            mean=mean, precision=precision, n_samples=n_samples, max_iter=max_iter
        )
        estimate_ = self._fit_frechet_mean(sample)

        self.assertAllClose(estimate_, expected, atol=atol)


class HypersphereIntrinsicTestCase(ManifoldTestCase):
    # TODO: update after refactoring

    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = HypersphereIntrinsicRandomDataGenerator(self.space)

    @pytest.mark.random
    def test_random_point_belongs(self, n_points, atol):
        point = self.space.random_point(n_points)
        expected = gs.ones(n_points, dtype=bool)

        res = _belongs_intrinsic(self.space, point)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    @pytest.mark.shape
    def test_random_point_shape(self, n_points):
        point = self.data_generator.random_point(n_points)

        expected_ndim = self.space.point_ndim + int(n_points > 1)
        self.assertEqual(gs.ndim(point), expected_ndim)

        self.assertAllEqual(gs.shape(point)[-self.space.point_ndim :], (self.space.dim))

        if n_points > 1:
            self.assertEqual(gs.shape(point)[0], n_points)

    def test_is_tangent(self, vector, base_point, expected, atol):
        res = _is_tangent_intrinsic(self.space, vector, base_point, atol=atol)
        self.assertAllEqual(res, expected)

    @pytest.mark.vec
    def test_is_tangent_vec(self, n_reps, atol):
        point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(point)

        res = _is_tangent_intrinsic(self.space, tangent_vec, point)

        vec_data = generate_vectorization_data(
            data=[dict(vector=tangent_vec, base_point=point, expected=res, atol=atol)],
            arg_names=["vector", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)



"""Group actions."""

import math
from abc import ABC, abstractmethod

import geomstats.backend as gs
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.special_orthogonal import SpecialOrthogonal
from geomstats.vectorization import get_batch_shape


class GroupAction(ABC):
    """Base class for group action."""

    @abstractmethod
    def __call__(self, group_elem, point):
        """Perform action of a group element on a manifold point.

        Parameters
        ----------
        group_elem : array-like, shape=[..., *group.shape]
            The element of a group.
        point : array-like, shape=[..., *space.shape]
            A point on a manifold.

        Returns
        -------
        orbit_point : array-like
            A point on the orbit of point.
        """


class ComposeAction(GroupAction):
    """Action of a group on itself by composition.

    Parameters
    ----------
    group : LieGroup
    lie_algebra_repr : bool
        If True, group elements are represented by the vector
        representation of Lie algebra elements.
        This is amenable to perform gradient-based optimizations
        on the Lie algebra.
    """

    def __init__(self, group, lie_algebra_repr=True):
        self._group = group
        self.lie_algebra_repr = lie_algebra_repr

    @property
    def group_elem_shape(self):
        """Shape of the group element representation."""
        if self.lie_algebra_repr:
            return (self._group.lie_algebra.dim,)
        return self._group.shape

    def __call__(self, group_elem, point):
        """Compose action of a group element on a point.

        Parameters
        ----------
        group_elem : array-like, shape=[..., *group_elem_shape]
            Group element in chosen representation.
        point : array-like, shape=[..., *group.shape]
            Point on the group.

        Returns
        -------
        orbit_point : array-like, shape=[..., *group.shape]
            A point in the orbit of point.
        """
        if self.lie_algebra_repr:
            algebra_elt = self._group.lie_algebra.matrix_representation(group_elem)
            group_elem = self._group.exp(algebra_elt)
        return self._group.compose(point, group_elem)


class SpecialOrthogonalComposeAction(ComposeAction):
    """Action of the special orthogonal group.

    Parameters
    ----------
    n : int
        Integer representing the shapes of the matrices : n x n.
    lie_algebra_repr : bool
        If True, group elements are represented by the vector
        representation of Lie algebra elements.
        This is amenable to perform gradient-based optimizations
        on the Lie algebra.
    """

    def __init__(self, n):
        super().__init__(SpecialOrthogonal(n, equip=False))


class CongruenceAction(GroupAction):
    """Congruence action."""

    def __call__(self, group_elem, point):
        """Congruence action of a group element on a matrix.

        Parameters
        ----------
        group_elem : array-like, shape=[..., n, n]
            The element of a group.
        point : array-like, shape=[..., n, n]
            A point on a manifold.

        Returns
        -------
        orbit_point : array-like, shape=[..., n, n]
            A point on the orbit of point.
        """
        return Matrices.mul(group_elem, point, Matrices.transpose(group_elem))


class PermutationAction(CongruenceAction):
    """Congruence action of the permutation group on matrices."""

    def __call__(self, group_elem, point):
        """Congruence action of a group element on a matrix.

        Parameters
        ----------
        group_elem : array-like, shape=[..., n]
            Permutations where in position i we have the value j meaning
            the node i should be permuted with node j.
        point : array-like, shape=[..., n, n]
            A point on a manifold.

        Returns
        -------
        orbit_point : array-like, shape=[..., n, n]
            A point on the orbit of point.
        """
        perm_mat = permutation_matrix_from_vector(group_elem, dtype=point.dtype)
        return super().__call__(perm_mat, point)


class RowPermutationAction(GroupAction):
    """Action of the permutation group on matrices by multiplication."""

    def __call__(self, group_elem, point):
        """Permutation action applied to matrix.

        Parameters
        ----------
        group_elem: array-like, shape=[..., n]
            Permutations where in position i we have the value j meaning
            the node i should be permuted with node j.
        point : array-like, shape=[..., n, n]
            Matrices to be permuted.

        Returns
        -------
        permuted_point : array-like, shape=[..., n, n]
            Permuted matrices.
        """
        perm_mat = permutation_matrix_from_vector(group_elem, dtype=point.dtype)
        return gs.matmul(Matrices.transpose(perm_mat), point)


def permutation_matrix_from_vector(group_elem, dtype=gs.int64):
    """Transform a permutation vector into a matrix.

    Parameters
    ----------
    group_elem : array-like, shape=[..., n]
        The element of a group.
    dtype: gs.dtype
        Array dtype

    Returns
    -------
    mat_group_element : array-like, shape=[..., n, n]
        Matrix representation of group element.
    """
    batch_shape = get_batch_shape(1, group_elem)
    n = group_elem.shape[-1]

    if batch_shape:
        indices = gs.array(
            [
                (k, i, j)
                for (k, group_elem_) in enumerate(group_elem)
                for (i, j) in zip(range(n), group_elem_)
            ]
        )
    else:
        indices = gs.array(list(zip(range(n), group_elem)))

    return gs.array_from_sparse(
        data=gs.ones(math.prod(batch_shape) * n, dtype=dtype),
        indices=indices,
        target_shape=batch_shape + (n, n),
    )



"""Visualization for Geometric Statistics."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs
from geomstats.geometry.special_orthogonal import SpecialOrthogonal

SO3_GROUP = SpecialOrthogonal(n=3, point_type="vector")


AX_SCALE = 1.2


class Arrow3D:
    """An arrow in 3d, i.e. a point and a vector."""

    def __init__(self, point, vector):
        self.point = point
        self.vector = vector

    def draw(self, ax, **quiver_kwargs):
        """Draw the arrow in 3D plot."""
        ax.quiver(
            self.point[0],
            self.point[1],
            self.point[2],
            self.vector[0],
            self.vector[1],
            self.vector[2],
            **quiver_kwargs,
        )


class Trihedron:
    """A trihedron, i.e. 3 Arrow3Ds at the same point."""

    def __init__(self, point, vec_1, vec_2, vec_3):
        self.arrow_1 = Arrow3D(point, vec_1)
        self.arrow_2 = Arrow3D(point, vec_2)
        self.arrow_3 = Arrow3D(point, vec_3)

    def draw(self, ax, **arrow_draw_kwargs):
        """Draw the trihedron by drawing its 3 Arrow3Ds.

        Arrows are drawn is order using green, red, and blue
        to show the trihedron's orientation.
        """
        if "color" in arrow_draw_kwargs:
            self.arrow_1.draw(ax, **arrow_draw_kwargs)
            self.arrow_2.draw(ax, **arrow_draw_kwargs)
            self.arrow_3.draw(ax, **arrow_draw_kwargs)
        else:
            blue = "#1f77b4"
            orange = "#ff7f0e"
            green = "#2ca02c"
            self.arrow_1.draw(ax, color=blue, **arrow_draw_kwargs)
            self.arrow_2.draw(ax, color=orange, **arrow_draw_kwargs)
            self.arrow_3.draw(ax, color=green, **arrow_draw_kwargs)

    def plot(self, points, ax=None, space=None, **point_draw_kwargs):
        if space == "SE3_GROUP":
            ax_s = AX_SCALE * gs.amax(gs.abs(points[:, 3:6]))
        elif space == "SO3_GROUP":
            ax_s = AX_SCALE * gs.amax(gs.abs(points[:, :3]))
        ax_s = float(ax_s)
        bounds = (-ax_s, ax_s)
        plt.setp(
            ax,
            xlim=bounds,
            ylim=bounds,
            zlim=bounds,
            xlabel="X",
            ylabel="Y",
            zlabel="Z",
        )
        trihedrons = convert_to_trihedron(points, space=space)
        for t in trihedrons:
            t.draw(ax, **point_draw_kwargs)


def convert_to_trihedron(point, space=None):
    """Transform a rigid point into a trihedron.

    Transform a rigid point into a trihedron such that:

    - the trihedron's base point is the translation of the origin
      of R^3 by the translation part of point,
    - the trihedron's orientation is the rotation of the canonical basis
      of R^3 by the rotation part of point.
    """
    point = gs.to_ndarray(point, to_ndim=2)
    n_points, _ = point.shape

    dim_rotations = SO3_GROUP.dim

    if space == "SE3_GROUP":
        rot_vec = point[:, :dim_rotations]
        translation = point[:, dim_rotations:]
    elif space == "SO3_GROUP":
        rot_vec = point
        translation = gs.zeros((n_points, 3))
    else:
        raise NotImplementedError(
            "Trihedrons are only implemented for SO(3) and SE(3)."
        )

    rot_mat = SO3_GROUP.matrix_from_rotation_vector(rot_vec)
    rot_mat = SO3_GROUP.projection(rot_mat)
    basis_vec_1 = gs.array([1.0, 0.0, 0.0])
    basis_vec_2 = gs.array([0.0, 1.0, 0.0])
    basis_vec_3 = gs.array([0.0, 0.0, 1.0])

    trihedrons = []
    for i in range(n_points):
        trihedron_vec_1 = gs.dot(rot_mat[i], basis_vec_1)
        trihedron_vec_2 = gs.dot(rot_mat[i], basis_vec_2)
        trihedron_vec_3 = gs.dot(rot_mat[i], basis_vec_3)
        trihedron = Trihedron(
            translation[i], trihedron_vec_1, trihedron_vec_2, trihedron_vec_3
        )
        trihedrons.append(trihedron)
    return trihedrons


def plot(points, ax=None, space=None, **point_draw_kwargs):
    """Plot trihedrons."""
    ax_s = AX_SCALE * gs.amax(gs.abs(points[:, :3]))
    ax_s = float(ax_s)
    bounds = (-ax_s, ax_s)
    plt.setp(
        ax,
        xlim=bounds,
        ylim=bounds,
        zlim=bounds,
        xlabel="X",
        ylabel="Y",
        zlabel="Z",
    )
    trihedrons = convert_to_trihedron(points, space=space)
    for t in trihedrons:
        t.draw(ax, **point_draw_kwargs)



from geomstats.test.test_case import TestCase


class RadialKernelFunctionsTestCase(TestCase):
    def test_kernel(self, kernel, distance, bandwidth, expected, atol):
        res = kernel(distance, bandwidth)
        self.assertAllClose(res, expected, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.exceptions import AutodiffNotImplementedError
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data


class NumpyRaisesTestCase(TestCase):
    def test_raises(self, autodiff_func):
        with pytest.raises(AutodiffNotImplementedError):
            autodiff_func(self.dummy_func)


class AutodiffTestCase(TestCase):
    def test_value_and_grad(
        self,
        func,
        inputs,
        expected_value,
        expected_grad,
        atol,
        argnums=0,
        point_ndims=1,
    ):
        value_and_grad = gs.autodiff.value_and_grad(
            func, argnums=argnums, point_ndims=point_ndims
        )
        value, grad = value_and_grad(*inputs)
        self.assertAllClose(value, expected_value, atol=atol)

        if isinstance(argnums, int):
            self.assertAllClose(grad, expected_grad, atol=atol)
        else:
            self.assertEqual(len(expected_grad), len(argnums))
            for grad_, expected_grad_ in zip(grad, expected_grad):
                self.assertAllClose(grad_, expected_grad_, atol=atol)

    def _test_value_and_grad(
        self,
        func,
        x,
        expected_value,
        expected_grad_x,
        atol,
        argnums=0,
        point_ndims=1,
        y=None,
        expected_grad_y=None,
    ):
        """Auxiliar func for vectorization test."""
        inputs = (x,) if y is None else (x, y)
        if isinstance(argnums, int):
            expected_grad = expected_grad_x if argnums == 0 else expected_grad_y
        elif len(argnums) == 1:
            argnum = argnums[0]
            expected_grad = (expected_grad_x if argnum == 0 else expected_grad_y,)
        else:
            expected_grad = (expected_grad_x, expected_grad_y)

        return self.test_value_and_grad(
            func,
            inputs,
            expected_value,
            expected_grad,
            atol,
            argnums=argnums,
            point_ndims=point_ndims,
        )

    @pytest.mark.vec
    def test_value_and_grad_vec(
        self,
        n_reps,
        func,
        input_shape_x,
        atol,
        argnums=0,
        point_ndims=1,
        input_shape_y=None,
    ):
        has_y = input_shape_y is not None

        inputs = [gs.random.rand(*input_shape_x)]
        if has_y:
            inputs.append(gs.random.rand(*input_shape_y))

        expected_value, expected_grad = gs.autodiff.value_and_grad(
            func,
            argnums=argnums,
            point_ndims=point_ndims,
        )(*inputs)

        argnums_ = (argnums,) if isinstance(argnums, int) else argnums
        expected_grad = (
            (expected_grad,) if not isinstance(expected_grad, tuple) else expected_grad
        )

        arg_names = ["x"]
        if has_y:
            arg_names.append("y")
        expected_names = ["expected_value"]

        expected_grad_ = []
        k = 0
        map_index_varname = {0: "x", 1: "y"}
        for index in range(2):
            if index not in argnums_:
                expected_grad_.append(None)
                continue

            expected_names.append(f"expected_grad_{map_index_varname[index]}")
            expected_grad_.append(expected_grad[k])
            k += 1

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    func=func,
                    x=inputs[0],
                    y=inputs[1] if input_shape_y is not None else None,
                    expected_value=expected_value,
                    expected_grad_x=expected_grad_[0],
                    expected_grad_y=expected_grad_[1],
                    argnums=argnums,
                    point_ndims=point_ndims,
                    atol=atol,
                )
            ],
            arg_names=arg_names,
            expected_name=expected_names,
            n_reps=n_reps,
        )

        self._test_vectorization(vec_data, test_fnc_name="_test_value_and_grad")


class MetricDistGradTestCase(TestCase):
    @pytest.mark.random
    def test_value_and_grad_sdist(self, n_points, atol):
        point_a = self.space.random_point(n_points)
        point_b = self.space.random_point(n_points)

        value_and_grad = gs.autodiff.value_and_grad(
            lambda v: self.space.metric.squared_dist(v, point_b),
            point_ndims=(self.space.point_ndim,),
        )
        sdist, sdist_grad = value_and_grad(point_a)

        expected_sdist = self.space.metric.squared_dist(point_a, point_b)

        tangent_vec = self.space.metric.log(point_b, point_a)
        expected_sdist_grad = -2 * tangent_vec

        self.assertAllClose(sdist, expected_sdist, atol=atol)
        self.assertAllClose(sdist_grad, expected_sdist_grad, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.test.test_case import TestCase
from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class RiemannianEMTestCase(BaseEstimatorTestCase):
    @pytest.mark.random
    def test_estimate_belongs(self, n_samples, atol):
        X = self.data_generator.random_point(n_points=n_samples)
        means = self.estimator.fit(X).means_
        belongs = self.estimator.space.belongs(means, atol=atol)
        self.assertAllEqual(belongs, gs.ones(self.estimator.n_gaussians, dtype=bool))

    @pytest.mark.random
    def test_fit_coefficients_and_variances_bounds(self, n_samples, atol):
        X = self.data_generator.random_point(n_points=n_samples)
        self.estimator.fit(X)

        coefficients = self.estimator.mixture_coefficients_
        variances = self.estimator.variances_

        self.assertTrue(
            (coefficients < 1 + atol).all() and (coefficients > -atol).all()
        )
        self.assertTrue((variances < 1 + atol).all() and (variances > -atol).all())


class GaussianMixtureModelTestCase(TestCase):
    def test_normalization_factor_init(
        self, expected_normalization_factor_var, expected_phi_inv_var, atol
    ):
        self.assertAllClose(
            self.model.normalization_factor_var[4],
            expected_normalization_factor_var,
            atol=atol,
        )
        self.assertAllClose(self.model.phi_inv_var[3], expected_phi_inv_var, atol=atol)

    def test_normalization_factor(self, expected_norm_factor, atol):
        norm_factor = self.model._compute_normalization_factor()
        self.assertAllClose(norm_factor, expected_norm_factor, atol=atol)

    def test_metric_normalization_factor(self, expected_norm_factor, atol):
        norm_factor = self.space.metric.normalization_factor(self.model.variances)
        self.assertAllClose(norm_factor, expected_norm_factor, atol=atol)

    def test_metric_norm_factor_gradient(
        self, expected_norm_factor, expected_norm_factor_gradient, atol
    ):
        (
            norm_factor,
            norm_factor_gradient,
        ) = self.model.space.metric.norm_factor_gradient(self.model.variances)

        self.assertAllClose(norm_factor, expected_norm_factor, atol=atol)
        self.assertAllClose(
            norm_factor_gradient, expected_norm_factor_gradient, atol=atol
        )

    def test_compute_variance_from_index(self, weighted_distances, expected_var, atol):
        var = self.model.compute_variance_from_index(weighted_distances)
        self.assertAllClose(var, expected_var, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class Model:
    # TODO: bring generation with noise
    def __init__(self, space, intercept, coef):
        self.space = space
        self.intercept = intercept
        self.coef = coef

    @property
    def param(self):
        return gs.vstack([self.intercept, self.coef])

    def generate(self, n_samples):
        time = gs.random.rand(n_samples)
        time = time - gs.mean(time)

        y = self.space.metric.exp(
            gs.einsum("n,...->n...", time, self.coef), self.intercept
        )
        return time, y


class GeodesicRegressionRandomDataGenerator(RandomDataGenerator):
    def random_model(self):
        intercept_true = self.random_point()
        coef_true = self.random_tangent_vec(intercept_true)

        return Model(self.space, intercept_true, coef_true)

    def random_X(self, n_samples):
        time = gs.random.rand(n_samples)
        time = time - gs.mean(time)
        return time

    def random_dataset(self, n_samples):
        # TODO: distinction between noisy and noiseless?
        model = self.random_model()
        X, y = model.generate(n_samples)
        return X, y


class GeodesicRegressionTestCase(BaseEstimatorTestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = GeodesicRegressionRandomDataGenerator(
                self.estimator.space
            )

    @pytest.mark.random
    def test_loss_at_true_is_zero(self, n_samples, atol):
        """Test that the loss is 0 at the true parameters."""
        model = self.data_generator.random_model()
        X, y = model.generate(n_samples)

        loss = self.estimator._loss(X, y, model.param)
        self.assertAllClose(loss, 0.0, atol=atol)

    @pytest.mark.random
    def test_param_belongs_and_is_tangent(self, n_samples, atol):
        space = self.estimator.space
        X, y = self.data_generator.random_dataset(n_samples)

        self.estimator.fit(X, y)

        intercept = self.estimator.intercept_
        coef = self.estimator.coef_

        belongs = space.belongs(intercept, atol=atol)
        self.assertTrue(belongs)

        is_tangent = space.is_tangent(coef, intercept, atol=atol)
        self.assertTrue(is_tangent)

    @pytest.mark.random
    def test_predict_and_score(self, n_samples, atol):
        X, y = self.data_generator.random_dataset(n_samples)  # assumes noiseless

        self.estimator.fit(X, y)

        score = self.estimator.score(X, y)
        self.assertTrue(gs.isclose(score, 1.0, atol=atol), msg=f"score: {score}")



import pytest

import geomstats.backend as gs
from geomstats.test.random import NFoldManifoldRandomDataGenerator, RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test_cases.geometry.manifold import ManifoldTestCase
from geomstats.test_cases.geometry.mixins import ProjectionTestCaseMixins
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase


class NFoldManifoldTestCase(ProjectionTestCaseMixins, ManifoldTestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = NFoldManifoldRandomDataGenerator(self.space)


class NFoldMetricTestCase(RiemannianMetricTestCase):
    pass


class NFoldMetricScalesTestCase(TestCase):
    def setup_method(self):
        if self.space.n_copies != 1:
            raise ValueError("Test case works only for n_copies=1")

        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.other_space)

    @pytest.mark.random
    def test_inner_product(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        axis = -(self.other_space.point_ndim + 1)
        tangent_vec_a_ = gs.expand_dims(tangent_vec_a, axis=axis)
        tangent_vec_b_ = gs.expand_dims(tangent_vec_b, axis=axis)
        base_point_ = gs.expand_dims(base_point, axis=axis)

        res = self.space.metric.inner_product(
            tangent_vec_a_, tangent_vec_b_, base_point_
        )
        other_res = self.other_space.metric.inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(res, other_res, atol=atol)



import math

import pytest

import geomstats.backend as gs
from geomstats.test.random import get_random_times
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.vectorization import get_batch_shape


class ProjectionTestCaseMixins:
    def test_projection(self, point, expected, atol):
        proj_point = self.space.projection(point)
        self.assertAllClose(proj_point, expected, atol=atol)

    @pytest.mark.vec
    def test_projection_vec(self, n_reps, atol):
        point = self.data_generator.point_to_project()
        expected = self.space.projection(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_projection_belongs(self, n_points, atol):
        """Check projection belongs to manifold.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point = self.data_generator.point_to_project(n_points)
        proj_point = self.space.projection(point)
        expected = gs.ones(n_points, dtype=bool)

        self.test_belongs(proj_point, expected, atol)


class GroupExpTestCaseMixins:
    def test_exp(self, tangent_vec, base_point, expected, atol):
        point = self.space.exp(tangent_vec, base_point)
        self.assertAllClose(point, expected, atol=atol)


class DistTestCaseMixins:
    def test_dist(self, point_a, point_b, expected, atol):
        res = self.space.metric.dist(point_a, point_b)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_dist_is_symmetric(self, n_points, atol):
        """Check distance is symmetric.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        dist_ab = self.space.metric.dist(point_a, point_b)
        dist_ba = self.space.metric.dist(point_b, point_a)

        self.assertAllClose(dist_ab, dist_ba, atol=atol)

    @pytest.mark.random
    def test_dist_is_positive(self, n_points, atol):
        """Check distance is positive.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        dist_ = self.space.metric.dist(point_a, point_b)
        res = gs.all(dist_ > -atol)
        self.assertTrue(res)

    @pytest.mark.random
    def test_dist_point_to_itself_is_zero(self, n_points, atol):
        """Check distance of a point to itself is zero.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point = self.data_generator.random_point(n_points)

        dist_ = self.space.metric.dist(point, point)

        expected_shape = (n_points,) if n_points > 1 else ()
        expected = gs.zeros(expected_shape)
        self.assertAllClose(dist_, expected, atol=atol)

    @pytest.mark.random
    def test_dist_triangle_inequality(self, n_points, atol):
        """Check distance satifies triangle inequality.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)
        point_c = self.data_generator.random_point(n_points)

        dist_ab = self.space.metric.dist(point_a, point_b)
        dist_bc = self.space.metric.dist(point_b, point_c)
        rhs = dist_ac = self.space.metric.dist(point_a, point_c)

        lhs = dist_ab + dist_bc
        res = gs.all(lhs + atol >= rhs)
        self.assertTrue(res, f"lhs: {lhs}, rhs: {dist_ac}, diff: {lhs-rhs}")


class GeodesicBVPTestCaseMixins:
    @pytest.mark.vec
    def test_geodesic_bvp_vec(self, n_reps, n_times, atol):
        initial_point, end_point = self.data_generator.random_point(2)
        time = get_random_times(n_times)

        expected = self.space.metric.geodesic(initial_point, end_point=end_point)(time)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    initial_point=initial_point,
                    end_point=end_point,
                    time=time,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["initial_point", "end_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data, test_fnc_name="test_geodesic")

    @pytest.mark.random
    def test_geodesic_boundary_points(self, n_points, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = gs.array([0.0, 1.0])

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)

        res = geod_func(time)
        expected = gs.stack(
            [initial_point, end_point], axis=-(self.space.point_ndim + 1)
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_geodesic_bvp_reverse(self, n_points, n_times, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = get_random_times(n_times)

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)
        geod_func_reverse = self.space.metric.geodesic(
            end_point, end_point=initial_point
        )

        res = geod_func(time)
        res_ = geod_func_reverse(1.0 - time)

        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_geodesic_bvp_belongs(self, n_points, n_times, atol):
        """Check geodesic belongs to manifold.

        This is for geodesics defined by the boundary value problem (bvp).

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = get_random_times(n_times)

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)
        points = geod_func(time)

        res = self.space.belongs(gs.reshape(points, (-1, *self.space.shape)), atol=atol)

        expected_shape = (
            math.prod(get_batch_shape(self.space.point_ndim, initial_point)) * n_times,
        )
        expected = gs.ones(expected_shape, dtype=bool)
        self.assertAllEqual(res, expected)



"""Discrete-path related machinery."""

import geomstats.backend as gs
from geomstats.numerics.finite_differences import forward_difference
from geomstats.numerics.interpolation import UniformUnitIntervalLinearInterpolator


class UniformlySampledPathEnergy:
    """Riemannian path energy of a uniformly-sampled path.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    """

    def __init__(self, space):
        self._space = space

    def __call__(self, path):
        """Compute Riemannian path energy.

        Parameters
        ----------
        path : array-like, shape=[..., n_times, *point_shape]
            Piecewise linear path.

        Returns
        -------
        energy : array-like, shape=[...,]
            Path energy.
        """
        return self.energy(path)

    def energy_per_time(self, path):
        """Compute Riemannian path enery per time.

        Parameters
        ----------
        path : array-like, shape=[..., n_times, *point_shape]
            Piecewise linear path.

        Returns
        -------
        energy : array-like, shape=[..., n_times - 1,]
            Stepwise path energy.
        """
        time_axis = -(self._space.point_ndim + 1)
        point_ndim_slc = tuple([slice(None)] * self._space.point_ndim)

        n_time = path.shape[time_axis]
        tangent_vecs = forward_difference(path, axis=time_axis)
        return self._space.metric.squared_norm(
            tangent_vecs,
            path[(..., slice(0, -1)) + point_ndim_slc],
        ) / (2 * (n_time - 1))

    def energy(self, path):
        """Compute Riemannian path energy.

        Parameters
        ----------
        path : array-like, shape=[..., n_times, *point_shape]
            Piecewise linear path.

        Returns
        -------
        energy : array-like, shape=[...,]
            Path energy.
        """
        return gs.sum(self.energy_per_time(path), axis=-1)


class UniformlySampledDiscretePath:
    """A uniformly-sampled discrete path.

    Parameters
    ----------
    path : array-like, [..., *point_shape]
    interpolator : Interpolator1D
    """

    def __init__(self, path, interpolator=None, **interpolator_kwargs):
        if interpolator is None:
            interpolator = UniformUnitIntervalLinearInterpolator(
                path, **interpolator_kwargs
            )
        self.interpolator = interpolator

    def __call__(self, t):
        """Interpolate path.

        Parameters
        ----------
        t : array-like, shape=[n_time]
            Interpolation time.

        Returns
        -------
        point : array-like, shape=[..., n_time, *point_shape]
        """
        if not gs.is_array(t):
            t = gs.array([t])

        if gs.ndim(t) == 0:
            t = gs.expand_dims(t, axis=0)
        return self.interpolator(t)



import pytest

import geomstats.backend as gs
from geomstats.geometry.matrices import Matrices
from geomstats.test.random import KendalShapeRandomDataGenerator, get_random_times
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import LevelSetTestCase
from geomstats.test_cases.geometry.fiber_bundle import FiberBundleTestCase
from geomstats.test_cases.geometry.quotient_metric import QuotientMetricTestCase
from geomstats.vectorization import get_batch_shape


def integrability_tensor_alt(total_space, tangent_vec_a, tangent_vec_b, base_point):
    r"""Compute the fundamental tensor A of the submersion.

    The fundamental tensor A is defined for tangent vectors of the total
    space by [ONeill]_ :math:`A_X Y = ver\nabla^M_{hor X} (hor Y)
    + hor \nabla^M_{hor X}( ver Y)` where :math:`hor,ver` are the
    horizontal and vertical projections.

    For the pre-shape space, we have closed-form expressions and the result
    does not depend on the vertical part of :math:`X`.

    (This is an alternative implementation.)

    Parameters
    ----------
    total_space : Manifold
        Total space with fiber bundle structure.
    tangent_vec_a : array-like, shape=[..., k_landmarks, ambient_dim]
        Tangent vector at `base_point`.
    tangent_vec_b : array-like, shape=[..., k_landmarks, ambient_dim]
        Tangent vector at `base_point`.
    base_point : array-like, shape=[..., k_landmarks, ambient_dim]
        Point of the total space.

    Returns
    -------
    vector : array-like, shape=[..., k_landmarks, ambient_dim]
        Tangent vector at `base_point`, result of the A tensor applied to
        `tangent_vec_a` and `tangent_vec_b`.

    References
    ----------
    .. [ONeill]  O’Neill, Barrett. The Fundamental Equations of a
        Submersion, Michigan Mathematical Journal 13, no. 4
        (December 1966): 459–69. https://doi.org/10.1307/mmj/1028999604.
    """
    # Only the horizontal part of a counts
    horizontal_a = total_space.fiber_bundle.horizontal_projection(
        tangent_vec_a, base_point
    )
    vertical_b, skew = total_space.fiber_bundle.vertical_projection(
        tangent_vec_b, base_point, return_skew=True
    )
    horizontal_b = tangent_vec_b - vertical_b

    # For the horizontal part of b
    transposed_point = Matrices.transpose(base_point)
    sigma = gs.matmul(transposed_point, base_point)
    alignment = gs.matmul(Matrices.transpose(horizontal_a), horizontal_b)
    right_term = alignment - Matrices.transpose(alignment)
    skew_hor = gs.linalg.solve_sylvester(sigma, sigma, right_term)
    vertical = -gs.matmul(base_point, skew_hor)

    # For the vertical part of b
    vert_part = -gs.matmul(horizontal_a, skew)
    tangent_vert = total_space.to_tangent(vert_part, base_point)
    horizontal_ = total_space.fiber_bundle.horizontal_projection(
        tangent_vert, base_point
    )

    return vertical + horizontal_


class PreShapeSpaceTestCase(LevelSetTestCase):
    def test_is_centered(self, point, expected, atol):
        res = self.space.is_centered(point, atol=atol)
        self.assertAllEqual(res, expected)

    @pytest.mark.vec
    def test_is_centered_vec(self, n_reps, atol):
        point = self.data_generator.random_point()
        expected = self.space.is_centered(point, atol=atol)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_center(self, point, expected, atol):
        res = self.space.center(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_center_vec(self, n_reps, atol):
        point = self.data_generator.random_point()
        expected = self.space.center(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_center_is_centered(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        res = self.space.center(point)

        expected = gs.ones(n_points, dtype=bool)
        self.test_is_centered(res, expected, atol)


class PreShapeBundleTestCase(FiberBundleTestCase):
    def _test_is_tangent(self, vector, base_point, expected, atol):
        res = self.total_space.is_tangent(vector, base_point, atol=atol)
        self.assertAllEqual(res, expected)

    def _get_horizontal_vec(self, base_point, return_tangent=False):
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        horizontal = self.total_space.fiber_bundle.horizontal_projection(
            tangent_vec,
            base_point,
        )
        if return_tangent:
            return tangent_vec, horizontal

        return horizontal

    def _get_vertical_vec(self, base_point, return_tangent=False):
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        vertical = self.total_space.fiber_bundle.vertical_projection(
            tangent_vec,
            base_point,
        )
        if return_tangent:
            return tangent_vec, vertical

        return vertical

    def _get_nabla_x_y(self, horizontal_vec_x, vec_y, base_point, horizontal=True):
        a_x_y = self.total_space.fiber_bundle.integrability_tensor(
            horizontal_vec_x, vec_y, base_point
        )

        dy = (
            self._get_horizontal_vec(base_point)
            if horizontal
            else self._get_vertical_vec(base_point)
        )

        return a_x_y + dy

    @pytest.mark.random
    def test_vertical_projection_correctness(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        transposed_point = Matrices.transpose(base_point)
        tmp_expected = gs.matmul(transposed_point, tangent_vec)
        expected = Matrices.transpose(tmp_expected) - tmp_expected

        vertical = self.total_space.fiber_bundle.vertical_projection(
            tangent_vec, base_point
        )
        tmp_result = gs.matmul(transposed_point, vertical)
        result = Matrices.transpose(tmp_result) - tmp_result

        self.assertAllClose(result, expected, atol=atol)

    @pytest.mark.random
    def test_horizontal_projection_correctness(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        horizontal = self.total_space.fiber_bundle.horizontal_projection(
            tangent_vec, base_point
        )

        result = gs.matmul(Matrices.transpose(base_point), horizontal)
        self.assertAllClose(result, Matrices.transpose(result), atol=atol)

    @pytest.mark.random
    def test_horizontal_projection_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        horizontal = self.total_space.fiber_bundle.horizontal_projection(
            tangent_vec, base_point
        )

        expected = gs.ones(n_points, dtype=bool)
        self._test_is_tangent(horizontal, base_point, expected, atol)

    @pytest.mark.random
    def test_alignment_is_symmetric(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        base_point = self.data_generator.random_point(n_points)

        aligned_point = self.total_space.fiber_bundle.align(point, base_point)
        alignment = Matrices.mul(Matrices.transpose(aligned_point), base_point)
        self.assertAllClose(alignment, Matrices.transpose(alignment), atol=atol)

    @pytest.mark.random
    def test_integrability_tensor_identity_1(self, n_points, atol):
        """Test integrability tensor identity.

        The integrability tensor A_X E is skew-symmetric with respect to the
        pre-shape metric, :math:`< A_X E, F> + <E, A_X F> = 0`. By
        polarization, this is equivalent to :math:`< A_X E, E> = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)

        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        result_ab = self.total_space.fiber_bundle.integrability_tensor(
            tangent_vec_a, tangent_vec_b, base_point
        )

        result = self.total_space.metric.inner_product(
            tangent_vec_b, result_ab, base_point
        )
        expected_shape = get_batch_shape(self.total_space.point_ndim, base_point)
        expected = gs.zeros(expected_shape)
        self.assertAllClose(result, expected, atol=atol)

    @pytest.mark.random
    def test_integrability_tensor_identity_2(self, n_points, atol):
        """Test integrability tensor identity.

        The integrability tensor is also alternating (:math:`A_X Y =
        - A_Y X`)  for horizontal vector fields :math:'X,Y',  and it is
        exchanging horizontal and vertical vector spaces.
        """
        base_point = self.data_generator.random_point(n_points)
        horizontal_a = self._get_horizontal_vec(base_point)
        tangent_vec_b, horizontal_b = self._get_horizontal_vec(
            base_point, return_tangent=True
        )

        result = self.total_space.fiber_bundle.integrability_tensor(
            horizontal_a, horizontal_b, base_point
        )
        expected = -self.total_space.fiber_bundle.integrability_tensor(
            horizontal_b, horizontal_a, base_point
        )
        self.assertAllClose(result, expected, atol=atol)

        is_vertical = self.total_space.fiber_bundle.is_vertical(
            result, base_point, atol=atol
        )
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(is_vertical, expected)

        vertical_b = tangent_vec_b - horizontal_b
        result = self.total_space.fiber_bundle.integrability_tensor(
            horizontal_a, vertical_b, base_point
        )
        is_horizontal = self.total_space.fiber_bundle.is_horizontal(
            result, base_point, atol=atol
        )
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(is_horizontal, expected)

    @pytest.mark.random
    def test_integrability_tensor_alt(self, n_points, atol):
        """Test if alternative and new implementation give the same result."""
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.integrability_tensor(
            tangent_vec_a, tangent_vec_b, base_point
        )
        alt_expected = integrability_tensor_alt(
            self.total_space, tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(expected, alt_expected, atol=atol)

    @pytest.mark.vec
    def test_integrability_tensor_derivative_vec(self, n_reps, atol):
        # TODO: need to be reviewed
        # TODO: can this be moved up?
        base_point = self.data_generator.random_point()
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        nabla_x_y = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_y, base_point)

        tangent_vec_e = self.data_generator.random_tangent_vec(base_point)
        nabla_x_e = self.data_generator.random_tangent_vec(base_point)

        (
            nabla_x_a_y_e,
            a_y_e,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            tangent_vec_e,
            nabla_x_e,
            base_point,
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    horizontal_vec_x=horizontal_vec_x,
                    horizontal_vec_y=horizontal_vec_y,
                    nabla_x_y=nabla_x_y,
                    tangent_vec_e=tangent_vec_e,
                    nabla_x_e=nabla_x_e,
                    base_point=base_point,
                    expected_nabla_x_a_y_e=nabla_x_a_y_e,
                    expected_a_y_e=a_y_e,
                    atol=atol,
                )
            ],
            arg_names=[
                "horizontal_vec_x",
                "horizontal_vec_y",
                "nabla_x_y",
                "tangent_vec_e",
                "nabla_x_e",
                "base_point",
            ],
            expected_name=["expected_nabla_x_a_y_e", "expected_a_y_e"],
            n_reps=n_reps,
            vectorization_type="basic",
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_integrability_tensor_derivative_is_alternate(self, n_points, atol):
        r"""Integrability tensor derivatives is alternate in pre-shape.

        For two horizontal vector fields :math:`X,Y` the integrability
        tensor (hence its derivatives) is alternate:
        :math:`\nabla_X ( A_Y Z + A_Z Y ) = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        horizontal_vec_z = self._get_horizontal_vec(base_point)
        nabla_x_y = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_y, base_point)
        nabla_x_z = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_z, base_point)

        (
            nabla_x_a_y_z,
            a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            horizontal_vec_z,
            nabla_x_z,
            base_point,
        )
        (
            nabla_x_a_z_y,
            a_z_y,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_z,
            nabla_x_z,
            horizontal_vec_y,
            nabla_x_y,
            base_point,
        )

        shape = (
            (n_points,) + self.total_space.shape
            if n_points > 1
            else self.total_space.shape
        )
        expected = gs.zeros(shape)

        result = nabla_x_a_y_z + nabla_x_a_z_y
        self.assertAllClose(result, expected, atol=atol)

        result = a_y_z + a_z_y
        self.assertAllClose(result, expected, atol=atol)

    @pytest.mark.random
    def test_integrability_tensor_derivative_is_skew_symmetric(self, n_points, atol):
        r"""Integrability tensor derivatives is alternate in pre-shape.

        For two horizontal vector fields :math:`X,Y` the integrability
        tensor (hence its derivatives) is alternate:
        :math:`\nabla_X ( A_Y Z + A_Z Y ) = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        horizontal_vec_z = self._get_horizontal_vec(base_point)
        nabla_x_y = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_y, base_point)
        nabla_x_z = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_z, base_point)
        ver_v = self.data_generator.random_tangent_vec(base_point)
        nabla_x_v = self.data_generator.random_tangent_vec(base_point)

        (
            nabla_x_a_y_z,
            a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            horizontal_vec_z,
            nabla_x_z,
            base_point,
        )

        (
            nabla_x_a_y_v,
            a_y_v,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            ver_v,
            nabla_x_v,
            base_point,
        )

        inner = self.total_space.metric.inner_product
        result = (
            inner(nabla_x_a_y_z, ver_v)
            + inner(a_y_z, nabla_x_v)
            + inner(nabla_x_a_y_v, horizontal_vec_z)
            + inner(a_y_v, nabla_x_z)
        )
        expected_shape = get_batch_shape(self.total_space.point_ndim, base_point)
        expected = gs.zeros(expected_shape)
        self.assertAllClose(result, expected, atol=atol)

    @pytest.mark.random
    def test_integrability_tensor_derivative_reverses_hor(self, n_points, atol):
        r"""Integrability tensor derivatives exchanges hor & ver in pre-shape.

        For :math:`X,Y,Z` horizontal and :math:`V,W` vertical, the
        integrability tensor (and thus its derivative) reverses horizontal
        and vertical subspaces: :math:`\nabla_X < A_Y Z, H > = 0`  and
        :math:`nabla_X < A_Y V, W > = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        horizontal_vec_z = self._get_horizontal_vec(base_point)
        nabla_x_y = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_y, base_point)
        nabla_x_z = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_z, base_point)

        (
            nabla_x_a_y_z,
            a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            horizontal_vec_z,
            nabla_x_z,
            base_point,
        )

        horizontal_vec_h = self._get_horizontal_vec(base_point)
        nabla_x_h = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_h, base_point)

        inner = self.total_space.metric.inner_product
        result = inner(nabla_x_a_y_z, horizontal_vec_h) + inner(a_y_z, nabla_x_h)
        expected_shape = get_batch_shape(self.total_space.point_ndim, base_point)
        expected = gs.zeros(expected_shape)
        self.assertAllClose(result, expected, atol=atol)

    @pytest.mark.random
    def test_integrability_tensor_derivative_reverses_ver(self, n_points, atol):
        r"""Integrability tensor derivatives exchanges hor & ver in pre-shape.

        For :math:`X,Y,Z` horizontal and :math:`V,W` vertical, the
        integrability tensor (and thus its derivative) reverses horizontal
        and vertical subspaces: :math:`\nabla_X < A_Y Z, H > = 0`  and
        :math:`nabla_X < A_Y V, W > = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        vertical_vec_z = self._get_vertical_vec(base_point)
        nabla_x_y = self._get_nabla_x_y(horizontal_vec_x, horizontal_vec_y, base_point)
        nabla_x_z = self._get_nabla_x_y(
            horizontal_vec_x, vertical_vec_z, base_point, horizontal=False
        )

        (
            nabla_x_a_y_z,
            a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            vertical_vec_z,
            nabla_x_z,
            base_point,
        )

        vertical_vec_w = self._get_vertical_vec(base_point)
        nabla_x_w = self._get_nabla_x_y(
            horizontal_vec_x, vertical_vec_w, base_point, horizontal=False
        )

        inner = self.total_space.metric.inner_product
        result = inner(nabla_x_a_y_z, vertical_vec_w) + inner(a_y_z, nabla_x_w)
        expected_shape = get_batch_shape(self.total_space.point_ndim, base_point)
        expected = gs.zeros(expected_shape)
        self.assertAllClose(result, expected, atol=atol)

    def test_integrability_tensor_derivative_parallel(
        self,
        horizontal_vec_x,
        horizontal_vec_y,
        horizontal_vec_z,
        base_point,
        expected_nabla_x_a_y_z,
        expected_a_y_z,
        atol,
    ):
        (
            nabla_x_a_y_z,
            a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative_parallel(
            horizontal_vec_x, horizontal_vec_y, horizontal_vec_z, base_point
        )
        self.assertAllClose(nabla_x_a_y_z, expected_nabla_x_a_y_z, atol=atol)
        self.assertAllClose(a_y_z, expected_a_y_z, atol=atol)

    @pytest.mark.vec
    def test_integrability_tensor_derivative_parallel_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        horizontal_vec_z = self._get_horizontal_vec(base_point)

        (
            expected_nabla_x_a_y_z,
            expected_a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative_parallel(
            horizontal_vec_x, horizontal_vec_y, horizontal_vec_z, base_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    horizontal_vec_x=horizontal_vec_x,
                    horizontal_vec_y=horizontal_vec_y,
                    horizontal_vec_z=horizontal_vec_z,
                    base_point=base_point,
                    expected_nabla_x_a_y_z=expected_nabla_x_a_y_z,
                    expected_a_y_z=expected_a_y_z,
                    atol=atol,
                )
            ],
            arg_names=["base_point"],
            expected_name=[
                "expected_nabla_x_a_y_z",
                "expected_a_y_z",
            ],
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_integrability_tensor_derivative_parallel_optimized(self, n_points, atol):
        """Test optimized integrability tensor derivatives in pre-shape space.

        Optimized version for quotient-parallel vector fields should equal
        the general implementation.
        """
        base_point = self.data_generator.random_point(n_points)
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        horizontal_vec_z = self._get_horizontal_vec(base_point)

        (
            nabla_x_a_y_z,
            a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative_parallel(
            horizontal_vec_x, horizontal_vec_y, horizontal_vec_z, base_point
        )

        a_x_y = self.total_space.fiber_bundle.integrability_tensor(
            horizontal_vec_x, horizontal_vec_y, base_point
        )
        a_x_z = self.total_space.fiber_bundle.integrability_tensor(
            horizontal_vec_x, horizontal_vec_z, base_point
        )
        (
            expected_nabla_x_a_y_z,
            expected_a_y_z,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            a_x_y,
            horizontal_vec_z,
            a_x_z,
            base_point,
        )

        self.assertAllClose(nabla_x_a_y_z, expected_nabla_x_a_y_z, atol=atol)
        self.assertAllClose(a_y_z, expected_a_y_z, atol=atol)

    def test_iterated_integrability_tensor_derivative_parallel(
        self,
        horizontal_vec_x,
        horizontal_vec_y,
        base_point,
        expected_nabla_x_a_y_v,
        expected_a_x_a_y_a_x_y,
        expected_nabla_x_v,
        expected_a_y_a_x_y,
        expected_vertical_vec_v,
        atol,
    ):
        (
            nabla_x_a_y_v,
            a_x_a_y_a_x_y,
            nabla_x_v,
            a_y_a_x_y,
            vertical_vec_v,
        ) = self.total_space.fiber_bundle.iterated_integrability_tensor_derivative_parallel(
            horizontal_vec_x, horizontal_vec_y, base_point
        )
        self.assertAllClose(nabla_x_a_y_v, expected_nabla_x_a_y_v, atol=atol)
        self.assertAllClose(a_x_a_y_a_x_y, expected_a_x_a_y_a_x_y, atol=atol)
        self.assertAllClose(nabla_x_v, expected_nabla_x_v, atol=atol)
        self.assertAllClose(a_y_a_x_y, expected_a_y_a_x_y, atol=atol)
        self.assertAllClose(vertical_vec_v, expected_vertical_vec_v, atol=atol)

    @pytest.mark.vec
    def test_iterated_integrability_tensor_derivative_parallel_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)
        (
            expected_nabla_x_a_y_v,
            expected_a_x_a_y_a_x_y,
            expected_nabla_x_v,
            expected_a_y_a_x_y,
            expected_vertical_vec_v,
        ) = self.total_space.fiber_bundle.iterated_integrability_tensor_derivative_parallel(
            horizontal_vec_x, horizontal_vec_y, base_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    horizontal_vec_x=horizontal_vec_x,
                    horizontal_vec_y=horizontal_vec_y,
                    base_point=base_point,
                    expected_nabla_x_a_y_v=expected_nabla_x_a_y_v,
                    expected_a_x_a_y_a_x_y=expected_a_x_a_y_a_x_y,
                    expected_nabla_x_v=expected_nabla_x_v,
                    expected_a_y_a_x_y=expected_a_y_a_x_y,
                    expected_vertical_vec_v=expected_vertical_vec_v,
                    atol=atol,
                )
            ],
            arg_names=["base_point"],
            expected_name=[
                "expected_nabla_x_a_y_v",
                "expected_a_x_a_y_a_x_y",
                "expected_nabla_x_v",
                "expected_a_y_a_x_y",
                "expected_vertical_vec_v",
            ],
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_iterated_integrability_tensor_derivative_parallel_optimized(
        self, n_points, atol
    ):
        """Test optimized iterated integrability tensor derivatives.

        The optimized version of the iterated integrability tensor
        :math:`A_X A_Y A_X Y`, computed with the horizontal lift of
        quotient-parallel vector fields extending the tangent vectors
        :math:`X,Y` of Kendall shape spaces (identified to horizontal vectors
        of the pre-shape space), is the recursive application of two general
        integrability tensor derivatives with proper derivatives.
        Intermediate computations returned are also verified.
        """
        base_point = self.data_generator.random_point(n_points)
        horizontal_vec_x = self._get_horizontal_vec(base_point)
        horizontal_vec_y = self._get_horizontal_vec(base_point)

        a_x_y = self.total_space.fiber_bundle.integrability_tensor(
            horizontal_vec_x, horizontal_vec_y, base_point
        )
        (
            nabla_x_v,
            a_x_y,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_x,
            gs.zeros_like(horizontal_vec_x),
            horizontal_vec_y,
            a_x_y,
            base_point,
        )

        (
            nabla_x_a_y_a_x_y,
            a_y_a_x_y,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x, horizontal_vec_y, a_x_y, a_x_y, nabla_x_v, base_point
        )

        a_x_a_y_a_x_y = self.total_space.fiber_bundle.integrability_tensor(
            horizontal_vec_x, a_y_a_x_y, base_point
        )

        (
            nabla_x_a_y_a_x_y_qp,
            a_x_a_y_a_x_y_qp,
            nabla_x_v_qp,
            a_y_a_x_y_qp,
            ver_v_qp,
        ) = self.total_space.fiber_bundle.iterated_integrability_tensor_derivative_parallel(
            horizontal_vec_x, horizontal_vec_y, base_point
        )
        self.assertAllClose(a_x_y, ver_v_qp, atol=atol)
        self.assertAllClose(a_y_a_x_y, a_y_a_x_y_qp, atol=atol)
        self.assertAllClose(nabla_x_v, nabla_x_v_qp, atol=atol)
        self.assertAllClose(a_x_a_y_a_x_y, a_x_a_y_a_x_y_qp, atol=atol)
        self.assertAllClose(nabla_x_a_y_a_x_y, nabla_x_a_y_a_x_y_qp, atol=atol)


class KendallShapeMetricTestCase(QuotientMetricTestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = KendalShapeRandomDataGenerator(self.total_space)
        super().setup_method()

    def _cmp_points(self, point, point_, atol):
        dists = self.space.metric.dist(point, point_)
        batch_shape = get_batch_shape(self.space.point_ndim, point, point_)

        self.assertAllClose(dists, gs.zeros(batch_shape), atol=atol)

    @pytest.mark.random
    def test_exp_after_log(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        tangent_vec = self.space.metric.log(end_point, base_point)
        end_point_ = self.space.metric.exp(tangent_vec, base_point)

        self._cmp_points(end_point_, end_point, atol=atol)

    @pytest.mark.random
    def test_log_after_exp(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        end_point = self.space.metric.exp(tangent_vec, base_point)
        tangent_vec_ = self.space.metric.log(end_point, base_point)

        end_point_ = self.space.metric.exp(tangent_vec_, base_point)

        self._cmp_points(end_point_, end_point, atol=atol)

    @pytest.mark.random
    def test_geodesic_bvp_reverse(self, n_points, n_times, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = get_random_times(n_times)

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)
        geod_func_reverse = self.space.metric.geodesic(
            end_point, end_point=initial_point
        )

        points = gs.reshape(geod_func(time), (-1, *self.space.shape))

        points_ = gs.reshape(geod_func_reverse(1.0 - time), (-1, *self.space.shape))

        self._cmp_points(points, points_, atol=atol)

    @pytest.mark.random
    def test_geodesic_boundary_points(self, n_points, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        time = gs.array([0.0, 1.0])

        geod_func = self.space.metric.geodesic(initial_point, end_point=end_point)

        points = gs.reshape(geod_func(time), (-1, *self.space.shape))
        expected_points = gs.reshape(
            gs.stack([initial_point, end_point], axis=-(self.space.point_ndim + 1)),
            (-1, *self.space.shape),
        )
        self._cmp_points(points, expected_points, atol=atol)

    @pytest.mark.random
    def test_curvature_is_skew_operator(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        res = self.space.metric.curvature(
            tangent_vec_a, tangent_vec_a, tangent_vec_b, base_point
        )

        batch_shape = (n_points,) if n_points > 1 else ()
        expected = gs.zeros(batch_shape + self.space.shape)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_curvature_bianchi_identity(self, n_points, atol):
        """First Bianchi identity on curvature in pre-shape space.

        :math:`R(X,Y)Z + R(Y,Z)X + R(Z,X)Y = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_c = self.data_generator.random_tangent_vec(base_point)

        curvature_1 = self.space.metric.curvature(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point
        )
        curvature_2 = self.space.metric.curvature(
            tangent_vec_b, tangent_vec_c, tangent_vec_a, base_point
        )
        curvature_3 = self.space.metric.curvature(
            tangent_vec_c, tangent_vec_a, tangent_vec_b, base_point
        )

        result = curvature_1 + curvature_2 + curvature_3

        batch_shape = (n_points,) if n_points > 1 else ()
        expected = gs.zeros(batch_shape + self.space.shape)
        self.assertAllClose(result, expected, atol=atol)

    @pytest.mark.random
    def test_sectional_curvature_lower_bound(self, n_points, atol):
        """Sectional curvature of Kendall shape space is larger than 1.

        The sectional curvature always increase by taking the quotient in a
        Riemannian submersion. Thus, it should larger in kendall shape space
        thane the sectional curvature of the pre-shape space which is 1 as it
        a hypersphere.
        The sectional curvature is computed here with the generic
        directional_curvature and sectional curvature methods.
        """
        base_point = self.data_generator.random_point(n_points)
        hor_a = self.data_generator.random_horizontal_vec(base_point)
        hor_b = self.data_generator.random_horizontal_vec(base_point)

        tidal_force = self.space.metric.directional_curvature(hor_a, hor_b, base_point)

        numerator = self.space.metric.inner_product(tidal_force, hor_a, base_point)
        denominator = (
            self.space.metric.inner_product(hor_a, hor_a, base_point)
            * self.space.metric.inner_product(hor_b, hor_b, base_point)
            - self.space.metric.inner_product(hor_a, hor_b, base_point) ** 2
        )

        condition = ~gs.isclose(denominator, 0.0, atol=atol)
        kappa = numerator[condition] / denominator[condition]
        kappa_direct = self.space.metric.sectional_curvature(hor_a, hor_b, base_point)[
            condition
        ]
        self.assertAllClose(kappa, kappa_direct)

        result = kappa > 1.0 - atol
        self.assertTrue(gs.all(result))

    @pytest.mark.random
    def test_curvature_derivative_bianchi_identity(self, n_points, atol):
        r"""2nd Bianchi identity on curvature derivative in kendall space.

        For any 3 tangent vectors horizontally lifted from kendall shape
        space to Kendall pre-shape space, :math:`(\nabla_X R)(Y, Z)
        + (\nabla_Y R)(Z,X) + (\nabla_Z R)(X, Y) = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        hor_x = self.data_generator.random_horizontal_vec(base_point)
        hor_y = self.data_generator.random_horizontal_vec(base_point)
        hor_z = self.data_generator.random_horizontal_vec(base_point)
        hor_h = self.data_generator.random_horizontal_vec(base_point)

        term_x = self.space.metric.curvature_derivative(
            hor_x, hor_y, hor_z, hor_h, base_point
        )
        term_y = self.space.metric.curvature_derivative(
            hor_y, hor_z, hor_x, hor_h, base_point
        )
        term_z = self.space.metric.curvature_derivative(
            hor_z, hor_x, hor_y, hor_h, base_point
        )

        result = term_x + term_y + term_z
        self.assertAllClose(result, gs.zeros_like(result), atol=atol)

    @pytest.mark.random
    def test_curvature_derivative_is_skew_operator(self, n_points, atol):
        r"""Derivative of a skew operator is skew.

        For any 3 tangent vectors horizontally lifted from kendall shape space
        to Kendall pre-shape space, :math:`(\nabla_X R)(Y,Y)Z = 0`.
        """
        base_point = self.data_generator.random_point(n_points)
        hor_x = self.data_generator.random_horizontal_vec(base_point)
        hor_y = self.data_generator.random_horizontal_vec(base_point)
        hor_z = self.data_generator.random_horizontal_vec(base_point)

        result = self.space.metric.curvature_derivative(
            hor_x, hor_y, hor_y, hor_z, base_point
        )
        self.assertAllClose(result, gs.zeros_like(result), atol=atol)

    @pytest.mark.random
    def test_directional_curvature_derivative_is_quadratic(self, n_points, atol):
        """Directional curvature derivative is quadratic in both variables."""
        base_point = self.data_generator.random_point(n_points)
        hor_x = self.data_generator.random_horizontal_vec(base_point)
        hor_y = self.data_generator.random_horizontal_vec(base_point)

        coef_x, coef_y = gs.random.uniform(size=(2,))

        res = self.space.metric.directional_curvature_derivative(
            coef_x * hor_x, coef_y * hor_y, base_point
        )
        expected = (
            coef_x**2
            * coef_y**2
            * self.space.metric.directional_curvature_derivative(
                hor_x, hor_y, base_point
            )
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_parallel_transport_ivp_transported_is_horizontal(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        direction = self.data_generator.random_tangent_vec(base_point)
        tangent_vec = self.data_generator.random_horizontal_vec(base_point)

        transported = self.space.metric.parallel_transport(
            tangent_vec, base_point, direction=direction
        )

        end_point = self.space.metric.exp(direction, base_point)

        res = self.total_space.fiber_bundle.is_horizontal(
            transported, end_point, atol=atol
        )

        expected_shape = get_batch_shape(self.space.point_ndim, base_point)
        expected = gs.ones(expected_shape, dtype=bool)

        self.assertAllEqual(res, expected)



"""Minkowski space.

Lead author: Nina Miolane.
"""

import geomstats.backend as gs
from geomstats.algebra_utils import from_vector_to_diagonal_matrix
from geomstats.geometry.euclidean import Euclidean, EuclideanMetric


class Minkowski(Euclidean):
    r"""Class for Minkowski space.

    This is the Euclidean space endowed with the inner-product of signature (
    dim-1, 1), i.e.
    .. math::
        ds^2 = - dx_1^2 + dx_2^2 + ... + dx_n^2

    Parameters
    ----------
    dim : int
       Dimension of Minkowski space.
    """

    def __new__(cls, dim, equip=True):
        """Instantiate a Minkowski space.

        This is an instance of the `Euclidean` class endowed with the
        `MinkowskiMetric`.
        """
        space = Euclidean(dim, equip=False)
        if equip:
            space.equip_with_metric(MinkowskiMetric)
        return space


class MinkowskiMetric(EuclideanMetric):
    """Class for the pseudo-Euclidean Minkowski metric."""

    def __init__(self, space):
        signature = (space.dim - 1, 1)

        q, p = signature
        diagonal = gs.array([-1.0] * p + [1.0] * q)
        metric_mat = from_vector_to_diagonal_matrix(diagonal)

        super().__init__(space=space, metric_matrix=metric_mat, signature=signature)



import pytest

import geomstats.backend as gs
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.general_linear import SquareMatrices
from geomstats.geometry.special_euclidean import (
    SpecialEuclidean,
    homogeneous_representation,
)
from geomstats.test.random import LieGroupVectorRandomDataGenerator
from geomstats.test.test_case import assert_allclose
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.lie_group import LieGroupTestCase
from geomstats.test_cases.geometry.mixins import ProjectionTestCaseMixins


def homogeneous_representation_test_case(
    rotation,
    translation,
    constant,
    expected,
    atol=gs.atol,
):
    out = homogeneous_representation(rotation, translation, constant)
    assert_allclose(out, expected, atol=atol)


def homogeneous_representation_vec_test_case(n, n_reps, atol):
    rotation = SquareMatrices(n).random_point()
    translation = Euclidean(n).random_point()
    constant = gs.array(1.0)

    expected = homogeneous_representation(rotation, translation, constant)

    vec_data = generate_vectorization_data(
        data=[
            dict(
                rotation=rotation,
                translation=translation,
                constant=constant,
                expected=expected,
                atol=atol,
            )
        ],
        arg_names=["rotation", "translation", "constant"],
        expected_name="expected",
        n_reps=n_reps,
    )
    for datum in vec_data:
        homogeneous_representation_test_case(**datum)


class SpecialEuclideanVectorsTestCase(ProjectionTestCaseMixins, LieGroupTestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = LieGroupVectorRandomDataGenerator(self.space)

    def test_matrix_from_vector(self, vec, expected, atol):
        mat = self.space.matrix_from_vector(vec)
        self.assertAllClose(mat, expected, atol=atol)

    @pytest.mark.vec
    def test_matrix_from_vector_vec(self, n_reps, atol):
        vec = self.data_generator.random_point()
        expected = self.space.matrix_from_vector(vec)

        vec_data = generate_vectorization_data(
            data=[dict(vec=vec, expected=expected, atol=atol)],
            arg_names=["vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_matrix_from_vector_belongs_to_matrices(self, n_points):
        point = self.data_generator.random_point(n_points)
        mat = self.space.matrix_from_vector(point)

        space = SpecialEuclidean(self.space.n, point_type="matrix")
        belongs = space.belongs(mat)

        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(expected, belongs)

    def test_regularize_tangent_vec(self, tangent_vec, base_point, expected, atol):
        regularized = self.space.regularize_tangent_vec(tangent_vec, base_point)
        self.assertAllClose(regularized, expected, atol=atol)

    def test_regularize_tangent_vec_at_identity(self, tangent_vec, expected, atol):
        regularized = self.space.regularize_tangent_vec_at_identity(tangent_vec)
        self.assertAllClose(regularized, expected, atol=atol)

    @pytest.mark.vec
    def test_regularize_tangent_vec_at_identity_vec(self, n_reps, atol):
        tangent_vec = self.data_generator.random_tangent_vec(self.space.identity)

        expected = self.space.regularize_tangent_vec_at_identity(tangent_vec)

        vec_data = generate_vectorization_data(
            data=[dict(tangent_vec=tangent_vec, expected=expected, atol=atol)],
            arg_names=["tangent_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)



"""Class for Stratified Spaces.

Lead authors: Anna Calissano and Jonas Lueg
"""

from abc import ABC, abstractmethod

import geomstats.backend as gs


class Point(ABC):
    """Class for points of a set."""

    @abstractmethod
    def equal(self, point, atol=gs.atol):
        """Check equality against another point.

        Parameters
        ----------
        point : Point or PointBatch
            Point to compare against point.
        atol : float

        Returns
        -------
        is_equal : array-like, shape=[...]
        """


class PointBatch(ABC, list):
    """Class for point batch."""

    def __getitem__(self, key):
        """Get item."""
        if isinstance(key, int):
            return list.__getitem__(self, key)
        return self.__class__(list.__getitem__(self, key))

    def equal(self, point, atol=gs.atol):
        """Check equality against another point.

        Parameters
        ----------
        point : Point or PointBatch
            Point to compare against point.
        atol : float
        """
        if isinstance(point, (list, tuple)):
            return gs.array(
                [
                    collection_point.equal(point_, atol)
                    for collection_point, point_ in zip(self, point)
                ]
            )

        return gs.array(
            [collection_point.equal(point, atol) for collection_point in self]
        )


class PointSet(ABC):
    r"""Class for a set of points of type Point."""

    def __init__(self, equip=True):
        if equip:
            self.equip_with_metric()

    def equip_with_metric(self, Metric=None, **metric_kwargs):
        """Equip manifold with Metric.

        Parameters
        ----------
        Metric : PointSetMetric object
            If None, default metric will be used.
        """
        if Metric is None:
            out = self.default_metric()
            if isinstance(out, tuple):
                Metric, kwargs = out
                kwargs.update(metric_kwargs)
                metric_kwargs = kwargs
            else:
                Metric = out

        self.metric = Metric(self, **metric_kwargs)

    @abstractmethod
    def belongs(self, point, atol=gs.atol):
        r"""Evaluate if a point belongs to the set.

        Parameters
        ----------
        point : Point or PointBatch
            Point to evaluate.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...]
            Boolean evaluating if point belongs to the set.
        """

    @abstractmethod
    def random_point(self, n_samples=1):
        r"""Sample random points on the PointSet.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.

        Returns
        -------
        samples : Point or PointBatch
            Points sampled on the PointSet.
        """


class PointSetMetric(ABC):
    r"""Class for the lenght spaces.

    Parameters
    ----------
    space : PointSet
        Set to equip with metric.
    """

    def __init__(self, space):
        self._space = space

    @abstractmethod
    def dist(self, point_a, point_b):
        """Distance between two points in the PointSet.

        Parameters
        ----------
        point_a: Point or PointBatch
            Point in the PointSet.
        point_b: Point or PointBatch
            Point in the PointSet.

        Returns
        -------
        distance : array-like, shape=[...]
            Distance.
        """

    @abstractmethod
    def geodesic(self, initial_point, end_point):
        """Compute the geodesic in the PointSet.

        Parameters
        ----------
        initial_point: Point or PointBatch
            Point in the PointSet.
        end_point: Point or PointBatch
            Point in the PointSet.

        Returns
        -------
        path : callable
            Time parameterized geodesic curve.
        """



"""(Oriented) varifolds related machinery.

General framework is introduced in [KCC2017]_.
See [CCGGR2020]_ for details about kernels.
Implementation is based in pykeops (https://www.kernel-operations.io/keops/).
In particular, see
https://www.kernel-operations.io/keops/_auto_tutorials/surface_registration/plot_LDDMM_Surface.html#data-attachment-term # noqa
for implementation details.

References
----------
.. [KCC2017] Irene Kaltenmark, Benjamin Charlier, and Nicolas Charon.
    “A General Framework for Curve and Surface Comparison and Registration
    With Oriented Varifolds,” 3346–55, 2017.
    https://openaccess.thecvf.com/content_cvpr_2017/html/Kaltenmark_A_General_Framework_CVPR_2017_paper.html.
.. [CCGGR2020] Nicolas Charon, Benjamin Charlier, Joan Glaunès, Pietro Gori, and Pierre Roussillon.
    “Fidelity Metrics between Curves and Surfaces: Currents, Varifolds, and Normal
    Cycles.” In Riemannian Geometric Statistics in Medical Image Analysis,
    edited by Xavier Pennec, Stefan Sommer, and Tom Fletcher, 441–77.
    Academic Press, 2020. https://doi.org/10.1016/B978-0-12-814725-2.00021-2
"""

import geomstats.backend as gs

if gs.__name__.endswith("pytorch"):
    from pykeops.torch import Vi, Vj
else:
    from pykeops.numpy import Vi, Vj


from geomstats._mesh import Surface


class SurfacesKernel:
    """A kernel on surfaces.

    Parameters
    ----------
    position_kernel : pykeops.LazyTensor
    tangent_kernel : pykeops.LazyTensor
    signal_kernel : pykeops.LazyTensor
    """

    def __init__(
        self,
        position_kernel=None,
        tangent_kernel=None,
        signal_kernel=None,
    ):
        reduction = 1.0

        self._has_position = False
        if position_kernel is not None:
            self._has_position = True
            reduction *= position_kernel

        self._has_tangent = False
        if tangent_kernel is not None:
            self._has_tangent = True
            reduction *= tangent_kernel

        self._has_signal = False
        if signal_kernel is not None:
            self._has_signal = True
            reduction *= signal_kernel

        area_b = Vj(reduction.new_variable_index(), 1)
        self._kernel = (reduction * area_b).sum_reduction(axis=1)

    def __call__(self, point_a, point_b):
        """Evaluate kernel.

        Parameters
        ----------
        point_a : Surface
        point_b : Surface

        Returns
        -------
        scalar : float
        """
        reduction_inputs = ()
        if self._has_position:
            reduction_inputs += (point_a.face_centroids, point_b.face_centroids)

        if self._has_tangent:
            reduction_inputs += (point_a.face_normals, point_b.face_normals)

        if self._has_signal:
            reduction_inputs += (point_a.signal, point_b.signal)

        reduction_inputs += (point_b.face_areas,)
        return gs.sum(self._kernel(*reduction_inputs) * point_a.face_areas)


def GaussianKernel(sigma=1.0, init_index=0, dim=3):
    r"""Gaussian kernel.

    .. math::

        K(x, y)=e^{-\|x-y\|^2 / \sigma^2}

    Generates the expression: `Exp(-SqDist(x,y)*a)`.

    Parameters
    ----------
    sigma : float
        Kernel parameter.
    init_index : int
        Index of first symbolic variable.
    dim : int
        Ambient dimension.
    """
    x, y = Vi(init_index, dim), Vj(init_index + 1, dim)
    gamma = 1 / (sigma * sigma)
    sdist = x.sqdist(y)
    return (-sdist * gamma).exp()


def CauchyKernel(sigma=1.0, init_index=0, dim=3):
    r"""Cauchy kernel.

    .. math::

        K(x, y)=\frac{1}{1+\|x-y\|^2 / \sigma^2}

    Generates the expression: `IntCst(1)/(IntCst(1)+SqDist(x,y)*a)`.

    Parameters
    ----------
    sigma : float
        Kernel parameter.
    init_index : int
        Index of first symbolic variable.
    dim : int
        Ambient dimension.
    """
    x, y = Vi(init_index, dim), Vj(init_index + 1, dim)
    gamma = 1 / (sigma * sigma)
    sdist = x.sqdist(y)
    return 1 / (1 + sdist * gamma)


def LinearKernel(init_index=0, dim=3):
    r"""Linear kernel.

    .. math::

        K(u, v) = \langle u, v \rangle

    Generates the expression: `(u|v)`.

    Parameters
    ----------
    init_index : int
        Index of first symbolic variable.
    dim : int
        Ambient dimension.
    """
    u, v = Vi(init_index, dim), Vj(init_index + 1, dim)
    return (u * v).sum()


def BinetKernel(init_index=0, dim=3):
    r"""Binet kernel.

    .. math::

        K(u, v) = \langle u, v \rangle^2

    Generates the expression: `Square((u|v))`.

    Parameters
    ----------
    init_index : int
        Index of first symbolic variable.
    dim : int
        Ambient dimension.
    """
    u, v = Vi(init_index, dim), Vj(init_index + 1, dim)
    return (u * v).sum() ** 2


def RestrictedGaussianKernel(sigma=1.0, oriented=False, init_index=0, dim=3):
    r"""Gaussian kernel restricted to the hypersphere.

    If unoriented:

    .. math::

        K(u, v)=e^{2 (\langle u, v \rangle ^2 - 1) / \sigma^2 }

    If oriented:

    .. math::

        K(u, v)=e^{2 (\langle u, v \rangle / - 1) - 1}

    Generates the expression:
    * oriented: `Exp(IntCst(2)*a*((u|v)-IntCst(1)))`
    * unoriented: `Exp(IntCst(2)*a*(Square((u|v))-IntCst(1)))`

    Parameters
    ----------
    sigma : float
        Kernel parameter.
    oriented : bool
        If False, uses squared inner product.
    init_index : int
        Index of first symbolic variable.
    dim : int
        Ambient dimension.
    """
    u, v = Vi(init_index, dim), Vj(init_index + 1, dim)
    b = 1 / (sigma * sigma)

    inner = (u * v).sum()
    if not oriented:
        inner = inner**2
    return (2 * b * (inner - 1)).exp()


class VarifoldMetric:
    """Varifold metric.

    Parameters
    ----------
    kernel : callable
    """

    def __init__(self, kernel=None):
        if kernel is None:
            position_kernel = GaussianKernel(sigma=1.0, init_index=0)
            tangent_kernel = BinetKernel(
                init_index=position_kernel.new_variable_index()
            )

            kernel = SurfacesKernel(position_kernel, tangent_kernel)

        self.kernel = kernel

    def scalar_product(self, point_a, point_b):
        """Scalar product.

        Parameters
        ----------
        point_a : Surface
            A point.
        point_b : Surface
            A point.

        Returns
        -------
        scalar : float
        """
        return self.kernel(point_a, point_b)

    def squared_dist(self, point_a, point_b):
        """Squared distance.

        Parameters
        ----------
        point_a : Surface
            A point.
        point_b : Surface
            A point.

        Returns
        -------
        scalar : float
        """
        return (
            self.kernel(point_a, point_a)
            - 2 * self.kernel(point_a, point_b)
            + self.kernel(point_b, point_b)
        )

    def dist(self, point_a, point_b):
        """Squared distance.

        Parameters
        ----------
        point_a : Surface
            A point.
        point_b : Surface
            A point.

        Returns
        -------
        scalar : float
        """
        sq_dist = self.squared_dist(point_a, point_b)
        return gs.sqrt(sq_dist)

    def loss(self, target_point, target_faces=None):
        """Loss with respected to target point.

        Parameters
        ----------
        point_a : Surface
            A point.
        target_faces : array-like, shape=[n_faces, 3]
            Combinatorial structure of target mesh.

        Returns
        -------
        squared_dist : callable
            `f(vertices) -> scalar`. Measures squared varifold distance
            between a point with `vertices` given wrt `target_faces`
            against `target_point`.
        """
        if target_faces is None:
            target_faces = target_point.faces

        kernel_target = self.kernel(target_point, target_point)

        def squared_dist(vertices):
            point = Surface(vertices, target_faces)
            return (
                kernel_target
                - 2 * self.kernel(target_point, point)
                + self.kernel(point, point)
            )

        return squared_dist



from geomstats.test_cases.geometry.base import VectorSpaceOpenSetTestCase
from geomstats.test_cases.geometry.matrices import MatricesMetricTestCase


class SPDMatricesTestCase(VectorSpaceOpenSetTestCase):
    pass


class SPDEuclideanMetricTestCase(MatricesMetricTestCase):
    def test_exp_domain(self, tangent_vec, base_point, expected, atol):
        res = self.space.metric.exp_domain(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)
